#re-write step
import numpy as np
import gc
import anndata
import pandas as pd
from torch.nn.modules.module import Module
import scanpy as sc
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn.neighbors import KernelDensity
from sklearn import cluster
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from scipy.stats import norm, entropy, multivariate_normal, gamma
from scipy import stats 
import torch
from pyro.primitives import param,deterministic
from torch.nn import functional as F
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from torch.distributions.gamma import Gamma
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from scipy.stats import entropy 
from scipy import stats
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
import threading
from scipy.stats import multivariate_normal
import pyro
import pyro.distributions as dist
from .distDistance import *
from scvi.module.base import PyroBaseModuleClass
import gc
import math
from sklearn.neighbors import kneighbors_graph

TTT = 0
from torch.distributions import constraints
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
)
from torch.nn.functional import softplus

from pyro.distributions import TorchDistribution, LogNormal, Poisson, Gamma, Weibull, Chi2
from pyro.distributions.util import broadcast_shape
import torch
import torch.nn.functional as F
from torch.distributions import constraints
from torch.distributions.distribution import Distribution
from torch.distributions.utils import broadcast_all, probs_to_logits, lazy_property, logits_to_probs

from numbers import Number

import torch
from torch.distributions import constraints
from torch.distributions.exp_family import ExponentialFamily
from torch.distributions.utils import broadcast_all
class ZeroInflatedDistribution(TorchDistribution):
    """
    Generic Zero Inflated distribution.

    This can be used directly or can be used as a base class as e.g. for
    :class:`ZeroInflatedPoisson` and :class:`ZeroInflatedNegativeBinomial`.

    :param TorchDistribution base_dist: the base distribution.
    :param torch.Tensor gate: probability of extra zeros given via a Bernoulli distribution.
    :param torch.Tensor gate_logits: logits of extra zeros given via a Bernoulli distribution.
    """

    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }

    def __init__(self, base_dist, *, gate=None, gate_logits=None, validate_args=None):
        if (gate is None) == (gate_logits is None):
            raise ValueError(
                "Either `gate` or `gate_logits` must be specified, but not both."
            )
        if gate is not None:
            batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
            self.gate = gate.expand(batch_shape)
        else:
            batch_shape = broadcast_shape(gate_logits.shape, base_dist.batch_shape)
            self.gate_logits = gate_logits.expand(batch_shape)
        if base_dist.event_shape:
            raise ValueError(
                "ZeroInflatedDistribution expected empty "
                "base_dist.event_shape but got {}".format(base_dist.event_shape)
            )

        self.base_dist = base_dist.expand(batch_shape)
        event_shape = torch.Size()
        
        super().__init__(batch_shape, event_shape, validate_args=False)

    

    @lazy_property
    def gate(self):
      
        return logits_to_probs(self.gate_logits, is_binary=True)

    @lazy_property
    def gate_logits(self):
        return probs_to_logits(self.gate, is_binary=True)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        
        if "gate" in self.__dict__:

            gate, value = broadcast_all(self.gate, value)
            temp_value = value.clone()
            temp_value[temp_value==0] = 1e-7
            log_prob = (-gate).log1p() + self.base_dist.log_prob(temp_value)
            log_prob = torch.where(value == 0, (gate).log(), log_prob)
        else:
            gate_logits, value = broadcast_all(self.gate_logits, value)
            temp_value = value.clone()
            temp_value[temp_value==0] = 1e-7
            temp_base_log_prob= self.base_dist.log_prob(temp_value)
            log_prob_minus_log_gate = -gate_logits + temp_base_log_prob#self.base_dist.log_prob(temp_value)
            log_gate = -softplus(-gate_logits)
            
            log_prob = log_prob_minus_log_gate + log_gate
            zero_log_prob = log_gate#-0.5#softplus(log_prob_minus_log_gate) + log_gate#+1
            log_prob = torch.where(value == 0, zero_log_prob, log_prob)
        return log_prob


    def sample(self, sample_shape=torch.Size()):
        
        shape = self._extended_shape(sample_shape)
        with torch.no_grad():
            mask = torch.bernoulli(self.gate.expand(shape)).bool()
            samples = self.base_dist.expand(shape).sample()
            samples = torch.where(mask, samples.new_zeros(()), samples)
        return samples


    @lazy_property
    def mean(self):
        return (1 - self.gate) * self.base_dist.mean

    @lazy_property
    def variance(self):
        return (1 - self.gate) * (
            self.base_dist.mean**2 + self.base_dist.variance
        ) - (self.mean) ** 2

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        gate = self.gate.expand(batch_shape) if "gate" in self.__dict__ else None
        gate_logits = (
            self.gate_logits.expand(batch_shape)
            if "gate_logits" in self.__dict__
            else None
        )
        base_dist = self.base_dist.expand(batch_shape)
        ZeroInflatedDistribution.__init__(
            new, base_dist, gate=gate, gate_logits=gate_logits, validate_args=False
        )
        new._validate_args = self._validate_args
        return new
class myZeroInflatedGamma(ZeroInflatedDistribution):
    """
    A Zero Inflated Normal distribution.

    :param total_count: non-negative number of negative Bernoulli trials.
    :type total_count: float or torch.Tensor
    :param torch.Tensor probs: Event probabilities of success in the half open interval [0, 1).
    :param torch.Tensor logits: Event log-odds for probabilities of success.
    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor gate_logits: logits of extra zeros.
    """

    arg_constraints = {
        'loc': constraints.real, 'scale': constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
   
    support = constraints.positive

    def __init__(
        self,
        loc,
        scale=None,
        gate=None,
        gate_logits=None,
        validate_args=None
    ):
        base_dist = Gamma(concentration=loc,rate=scale,validate_args=False)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )

class myZeroInflatedLogNormal(ZeroInflatedDistribution):
    """
    A Zero Inflated Normal distribution.

    :param total_count: non-negative number of negative Bernoulli trials.
    :type total_count: float or torch.Tensor
    :param torch.Tensor probs: Event probabilities of success in the half open interval [0, 1).
    :param torch.Tensor logits: Event log-odds for probabilities of success.
    :param torch.Tensor gate: probability of extra zeros.
    :param torch.Tensor gate_logits: logits of extra zeros.
    """

    arg_constraints = {
        'loc': constraints.real, 'scale': constraints.positive,
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
    }
   
    support = constraints.positive

    def __init__(
        self,
        loc,
        scale=None,
        gate=None,
        gate_logits=None,
        validate_args=None
    ):
        base_dist = LogNormal(loc=loc,scale=scale,validate_args=False)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
        )
        
    
class GraphConvolution(Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
#         self.fc = nn.Linear(in_features, out_features)
        self.out_features = out_features
#         self.weight = torch.nn.Parameter(torch.FloatTensor(in_features, out_features))
#         self.BN1 = nn.BatchNorm1d(out_features)
#         self.BN2 = nn.BatchNorm1d(out_features)
#         if bias:
#             self.bias = torch.nn.Parameter(torch.FloatTensor(out_features))
#         else:
#             self.register_parameter('bias', None)
        #self.reset_parameters()
#         self.softplus = nn.Sigmoid()
    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        # print(input.shape,self.weight.shape)
        # support = self.softplus(self.fc(input.float()))
        # support = self.BN2(support)
        
        # # print(support.shape)
        # support = torch.mm(input.float(), self.weight.float())
        # output = torch.spmm(adj, support)
        # print(torch.sum(input))

        support = torch.spmm(adj, input)
        # print(torch.sum(support))
        
        # output = self.fc(support)
        # output = torch.mm(support, self.weight.float())
        return support, support 
#         if self.bias is not None:
#             return support, support #+ self.bias
#         else:
#             return support,support

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'

class GraphEncoder(nn.Module):
    def __init__(self, in_dim, z_dim, hidden_dim):
        super().__init__()
        
        self.gc1 = GraphConvolution(in_dim, hidden_dim)
        # self.gc1 = GraphConvolution(in_dim, hidden_dim)
        # self.dropout = dropout
#         self.fc1 = nn.Linear(nhid, 512)
#         self.fc2 = nn.Linear(512, hidden_dim)
        #self.fc3 = nn.Linear(256, hidden_dim)
        self.fc21 = nn.Linear(hidden_dim, z_dim)
        self.fc22 = nn.Linear(hidden_dim, z_dim)
#         self.BN1 = nn.BatchNorm1d(hidden_dim)
        
        self.test = nn.Linear(in_dim, hidden_dim)
#         self.test1 = nn.Linear(hidden_dim, hidden_dim)
        self.softplus = nn.Softplus()

    def forward(self, x,adj,  start, end, test=False): 
        if test == False:
            y, x = self.gc1(x, adj)
            hidden = x.clone()
            # x = self.softplus(x)
            hidden1 = x
            # hidden=x[start:end,:] #temporarily remove to test full batch vs small batch
            # hidden = self.softplus(self.BN1(x)) #temporarily remove to test full batch vs small batch
            y=y[start:end,:]
            # hidden = self.softplus(self.test(x))
            # x = F.dropout(x, self.dropout, training=self.training)

            hidden1=x[start:end,:]
        else:
            y = x
            hidden1 = x.clone()
        
        # print(x)
        
        hidden1 = self.softplus(self.test(hidden1))
#         hidden1 = self.softplus(self.test1(hidden1))
#         hidden1=  self.softplus(self.test1(hidden1))
        # hidden = self.softplus(x)
        # hidden1 = self.softplus(self.fc1(x))
        # hidden1 = self.softplus(self.fc2(hidden1))
        #hidden1 = self.softplus(self.fc3(hidden1))
        # print(hidden)
        z_loc = self.fc21(hidden1)
        z_scale = torch.exp(torch.clamp(self.fc22(hidden1), -5, 5))
#         print(z_loc)
        return z_loc, z_scale,y,z_scale
class Discriminator(nn.Module):
    def __init__(self,input_dim,hidden_dim):
        super().__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, 1)
#         self.fc3 = nn.Linear(64, 1)
        self.ReLU6 = nn.Softplus()
        self.BN1 = nn.BatchNorm1d(hidden_dim)
        
        self.bce = nn.BCEWithLogitsLoss()
    def forward(self, x, y):
        ys = y
        xs = x
        xs = self.ReLU6(self.fc1(xs))#+1e-6

#         print(xs)
#         xs = self.ReLU6(self.fc2(xs))+1e-6
        xs = self.fc2(xs)

#         ys = torch.LongTensor(ys)
        return self.bce(xs,ys)
class NewDecoder(nn.Module):
    def __init__(self, in_dim, z_dim, hidden_dim):
        super().__init__()

        self.fc1 = nn.Linear(z_dim, hidden_dim)
#         self.fc2 = nn.Linear(512, hidden_dim)
#         self.test = nn.Linear(hidden_dim,hidden_dim)
#         self.fc3 = nn.Linear(512, hidden_dim)
        self.fc21 = nn.Linear(hidden_dim, in_dim)
        self.fc22 = nn.Linear(hidden_dim, in_dim)
        self.fc23 = nn.Linear(hidden_dim, in_dim)
#         self.fc24 = nn.Linear(hidden_dim, in_dim)
        self.softplus = nn.Softplus()
        self.ReLU6 = nn.ReLU6()
        self.ReLU = nn.ReLU()
        self.sigmoid = nn.Sigmoid()
    def forward(self, z):

        hidden =self.softplus(self.fc1(z))#+1e-7
        
#         loc =  torch.exp(torch.clamp(self.fc21(hidden), -5, 5))#+1e-6
        loc =  self.fc21(hidden) #mu
#         scale = torch.exp(torch.clamp(self.fc22(hidden), -5, 5))#+1e-6
        scale = self.softplus(self.fc22(hidden))
        dropout = self.sigmoid(torch.clamp(self.fc23(hidden),-3,3))
        return loc, scale, dropout
    
class VAE(PyroBaseModuleClass):
    def __init__(self,  n_input, n_latent,n_graph,beta,distribution): 
        super().__init__()
        self.n_latent = n_latent
        self.n_input = n_input
        self.n_graph = n_graph
        self.beta = beta
        self.distribution = distribution
       # in the init, we create the parameters of our elementary stochastic computation unit.
       
        # First, we setup the parameters of the generative model
        self.decoder = NewDecoder(self.n_input,self.n_latent,self.n_graph)
        self.log_theta = torch.nn.Parameter(torch.randn(self.n_input))
        self.gate_logits = torch.nn.Parameter(torch.randn(self.n_input))
        self.discriminator = Discriminator(self.n_input, self.n_latent)
        # Second, we setup the parameters of the variational distribution
        self.encoder = GraphEncoder(self.n_input, self.n_latent,self.n_graph)
        # self.gc = GraphConvolution(n_input, nhid)
        # self.softplus = nn.Sigmoid()
        
    def model(self, x,adj,batch,  start, end):
        # register PyTorch module `decoder` with Pyro p(z) p(x|z)
        pyro.module("decoder", self)
        with pyro.plate("data", x[start:end,:].shape[0]):

            # setup hyperparameters for prior p(z)
            z_loc = x.new_zeros(torch.Size((x[start:end,:].shape[0], self.n_latent)))
            z_scale = x.new_ones(torch.Size((x[start: end,:].shape[0], self.n_latent)))
            # z_p = Variable(torch.randn(64,128))
            # sample from prior (value will be sampled by guide when computing the ELBO)
            z = pyro.sample("latent", dist.Normal(z_loc, z_scale).to_event(1))
            # get the "normalized" mean of the negative binomial
            if self.distribution == 'ziln':
                dec_loc, dec_mu, dec_dropout  = self.decoder(z)
                scale = torch.exp(self.log_theta)
                loc = (dec_mu+1e-5).log()-scale.pow(2)/2
                x_dist = myZeroInflatedLogNormal(loc=loc, scale=scale, gate=dec_dropout)
                # score against actual counts
                rx=pyro.sample("obs", x_dist.to_event(1), obs=x[start:end,:])
                a = deterministic('recon',x_dist.mean)
            elif self.distribution == 'zinb':
                _, mu, dec_dropout  = self.decoder(z)
                theta = self.log_theta.exp()
                nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            
                x_dist = dist.ZeroInflatedNegativeBinomial(total_count=theta, logits=nb_logits,gate= dec_dropout, validate_args=False)
                # score against actual counts
                rx=pyro.sample("obs", x_dist.to_event(1), obs=x[start:end,:])
                a = deterministic('recon',x_dist.mean)
            elif self.distribution =='nb':
                _, mu, dec_dropout  = self.decoder(z)
                theta = self.log_theta.exp()
                nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            
                x_dist = dist.NegativeBinomial(total_count=theta, logits=nb_logits, validate_args=False)
                # score against actual counts
                rx=pyro.sample("obs", x_dist.to_event(1), obs=x[start:end,:])
                a = deterministic('recon',x_dist.mean)

            elif self.distribution =='zig':
                _, mu, dec_dropout = self.decode(z)
                scale = F.softplus(self.log_theta)#self.log_theta.exp()
                loc = mu*scale
                rx = myZeroInflatedGamma(loc=loc, scale=scale, gate=dec_dropout, validate_args=False)
                a = deterministic('recon',rx.mean)
            else:
                raise ValueError('distribution not supported')
            return rx
        
    def guide(self, x, adj, batch, start, end):
        # define the guide (i.e. variational distribution) q(z|x)
        pyro.module("encoder", self)
        
        with pyro.plate("data", x[start:end,:].shape[0]):
            # use the encoder to get the parameters used to define q(z|x)
#             x_ = torch.log(1 + x)
            [qz_m,qz_v,_,_] = self.encoder(x, adj,  start, end)
          
            #qz_v = self.var_encoder(x_)
            # sample the latent code z
            rz=pyro.sample("latent", dist.Normal(qz_m, qz_v).to_event(1))
            return rz
        
    def getZ(self, x,adj,batch, start, end,test=True):
        # encode image x
        # x = torch.log(1+x)
        z_loc, z_scale,gcnout,hidden = self.encoder(x,adj,  start, end,test=test)
        # sample in latent space
        # z = dist.Normal(z_loc, z_scale).sample()
        
        return z_loc+z_scale,z_loc, z_scale,gcnout,hidden

    
    # def generate(self,x, adj,batch, start, end):
    #     z_loc, z_scale,gcnout,hidden = self.encoder(x,adj,batch,  start, end,test=True)
    #     #z_loc = torch.zeros([cell,64]).to('cuda:0')+z_loc
    #     #z_scale = torch.zeros([cell,64]).to('cuda:0')+z_scale
    #     z = dist.Normal(z_loc, z_scale).sample()
    #     #z = z.mean(axis=0)
    #     dec_loc, dec_scale, dec_dropout = self.decoder(z_loc)

    #     return  dec_loc, dec_scale
    def generate1(self,x,adj,batch, start, end):
        # x = torch.log(1 + x)
        z_loc, z_scale,gcnout,hidden = self.encoder(x,adj,batch,  start, end,test=True)
        z = dist.Normal(z_loc, z_scale).sample()
        if self.distribution == 'ziln':
            dec_loc, dec_mu, dec_dropout  = self.decoder(z_loc+z_scale)
            scale = torch.exp(self.log_theta)
            loc = (dec_mu+1e-5).log()-scale.pow(2)/2
            x_dist = myZeroInflatedLogNormal(loc=loc, scale=scale, gate=dec_dropout)
            rx=x_dist.mean #-> expectation
        elif self.distribution == 'zinb':
            _, mu, dec_dropout  = self.decoder(z_loc+z_scale)
            theta = self.log_theta.exp()
            nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            x_dist = dist.ZeroInflatedNegativeBinomial(total_count=theta, logits=nb_logits,gate= dec_dropout, validate_args=False)
            rx=x_dist.mean
        elif self.distribution =='zig':
            _, mu, dec_dropout = self.decoder(z_loc+z_scale)
            scale = F.softplus(self.log_theta)
            loc = mu*scale
            x_dist = myZeroInflatedGamma(loc=loc, scale=scale, gate=dec_dropout, validate_args=False)
            rx = x_dist.mean
        # rx=x_dist.sample() #-> sample
        elif self.distribution =='nb':
            _, mu, dec_dropout  = self.decoder(z_loc+z_scale)
            theta = self.log_theta.exp()
            nb_logits = (mu+1e-4).log() - (theta+1e-4).log()
            x_dist = dist.NegativeBinomial(total_count=theta, logits=nb_logits, validate_args=False)
            rx=x_dist.mean
        else:
            raise ValueError('distribution not supported')
        return  rx#, None, None,None, None


import numpy as np
import gc
import anndata
import pandas as pd
import scanpy as sc
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn.neighbors import KernelDensity
from sklearn import cluster
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from scipy.stats import norm, entropy, multivariate_normal, gamma
from scipy import stats 
import torch
from torch.nn import functional as F
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from torch.distributions.gamma import Gamma
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from scipy.stats import entropy 
from scipy import stats
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
import threading
from scipy.stats import multivariate_normal
import gc
import os
from .distDistance import *
def nodesDistance(rep1,rep2,topgene1,topgene2):
    '''
    calculate the distance between two stages
    args: 
    rep1: the representation of clusters in stage 1
    rep2: the representation of clusters in stage 2
    topgene1: top 100 differential gene of clusters in stage 1
    topgene2: top 100 differential gene of clusters in stage 2
    return:
    distance: normalized distance of clusters between two stages
    '''
    distance = [[] for _ in range(len(rep2))]
    for i in range(len(rep2)):
        for j in range(len(rep1)):
            gaussiankl_1 = calculateKL(rep2[i],rep1[j])
            gaussiankl_2 = calculateKL(rep1[j],rep2[i])
            gaussiankl = (gaussiankl_1 + gaussiankl_2)/2
            similarityDE = getSimilarity(topgene2,topgene1,i,j)
            distance[i].append([gaussiankl , similarityDE])
    for i in range(len(distance)):
        distance[i] = normalizeDistance(distance[i])
    return distance
def connectNodes(distances):
    edges = []
    for i in range(len(distances)):
        leftend = np.argmin(distances[i])
        temp = distances[i].copy()
        pval = norm.cdf(distances[i][leftend],loc = np.mean(distances),scale = np.std(distances))
        #p*count/(count-idx)
        # q_val = pval * len(temp)/
        
        if pval < 0.01: #if pval < 0.01 can connect the two clusters across two stages
            edges.append([leftend,i])
    return edges

def buildEdges(stage1,stage2):
    '''
    calculate the distance between two stages and connect the clusters in two stages with smallest distance
    args: 
    stage1,
    stage2
    return:
    edges: the edges between two stages
    '''
    adata1 = sc.read_h5ad('./stagedata/%d.h5ad'%stage1)
    adata2 = sc.read_h5ad('./stagedata/%d.h5ad'%stage2)
    reps = np.load('./stagedata/rep.npy',allow_pickle=True)
    
    rep1 = reps[stage1]
    rep2 = reps[stage2]
    topgene1 = adata1.uns['topGene']
    topgene2 = adata2.uns['topGene']
    distance = nodesDistance(rep1,rep2,topgene1,topgene2)
    edges = connectNodes(distance)
    return edges
def buildEdges(stage1,stage2,midpath,iteration):
    '''
    calculate the distance between two stages and connect the clusters in two stages with smallest distance with midpath in iterative training
    args: 
    stage1,
    stage2
    return:
    edges: the edges between two stages
    '''
    # print(stage1,stage2)
    adata1 = sc.read_h5ad(os.path.join(midpath,str(iteration)+'/stagedata/%d.h5ad'%stage1))
    adata2 = sc.read_h5ad(os.path.join(midpath,str(iteration)+'/stagedata/%d.h5ad'%stage2))
    reps = np.load(os.path.join(midpath,str(iteration)+'/stagedata/rep.npy'),allow_pickle=True)
    
    rep1 = reps[stage1]
    rep2 = reps[stage2]
    topgene1 = adata1.uns['topGene']

    topgene2 = adata2.uns['topGene']
    distance = nodesDistance(rep1,rep2,topgene1,topgene2)
    edges = connectNodes(distance)
    return edges
def getandUpadateEdges(total_stage,midpath,iteration):
    '''
    get edges with midpath in iterative training
    '''
    edges = []
    for i in range(total_stage-1):
        edges.append(buildEdges(i,i+1,midpath,iteration))
    updateEdges(edges,midpath,iteration)
    # print('edges updated')
    return edges

def getEdges():
    '''
    get edges
    '''
    edges = []
    for i in range(3):
        edges.append(buildEdges(i,i+1))
    return edges
def updateEdges(edges,midpath,iteration):
    '''
    updata edges to the anndata database, calculate edges changes
    args:
    adata: anndata of database
    edges: edges from buildEdges()

    return: 
    adata: updated anndata of database
    '''
    newEdges = {}
    for i in range(len(edges)):
        newEdges[str(i)] = edges[i]
    #adata.uns['edges'] = newEdges
    f = open(os.path.join(midpath,str(iteration)+'/edges.txt'),'w')
    f.write(str(newEdges))
    f.close()

def reupdateAttributes(adata, stage, results):
    '''
    update gaussian and gamma rep, top 100 differential genes, cell types of clusters to anndata

    args: 
    adata: anndata of database
    results: [gaussian, gamma], a list contained top differential genes and cell types of clusters
    
    returns: 
    adata: updated anndata of database
    '''
    stageids = adata.obs[adata.obs['stage'] == stage].index.tolist()
    tempadata = adata[stageids]
    adata.uns['rep'][str(stage)] = results[0]
    adata.uns['topGene'][str(stage)] = results[1]
    adata.uns['clusterType'][str(stage)] = results[2]
    return adata
def averageNode(nodes,state):
    '''
    calculate the average gene expression of sibling nodes
    args: 
    nodes: number of sibling nodes
    state: the gene expression of each cluster in a certain stage
    
    return: 
    out: the average gene expression of sibling nodes
    '''
    out = 0
    for each in nodes:
        out+=state[each]
    return out/len(nodes)
def getClusterIdrem(paths,state):
    '''
    concatenate the average gene expression in a cluster tree. shape: [number of stages, number of gene]
    args: 
    paths: the collection of paths
    state: a list of average gene expression of each state
    
    return: 
    out: a list of gene expression of each cluster tree
    '''
    out = []
    for i,each in enumerate(paths.keys()):
        if len(paths[each]) == 4:
            stages = []
            for j, item in enumerate(paths[each]):
                stages.append(averageNode(item,state[j]))
            

            joint_matrix = np.concatenate((stages[0].reshape(-1,1), stages[1].reshape(-1,1), stages[2].reshape(-1,1), stages[3].reshape(-1,1)),axis =1)
            out.append(joint_matrix)
    return out


import gc
import os
import shutil
import pickle
import scanpy as sc
import subprocess
import numpy as np
from .afterall_utils import find_overlap_and_assign_direction,calculateDataPathwayOverlapGene,calculateTopPathwayGeneRanking
from .get_hcmarkers import get_dataset_hcmarkers
from .speedup_perturbation import perturbation
from .shuffle_progressionMarker_background import get_progressionmarker_background
from .getProgressionTopMarkers import runGetProgressionMarker_one_dist
class analyst():
    def __init__(self,data_path,iteration,target_dir=None,customized_drug=None,cmap_dir=None):
        self.adata = sc.read(data_path)
        self.data_folder = os.path.dirname(data_path)
        self.adata.uns = pickle.load(open(self.data_folder+'/org_attribute.pkl', 'rb'))
        self.customized_drug = customized_drug
        self.cmap_dir = cmap_dir
        self.iteration = iteration
        if target_dir is None:
            self.target_dir = './'+self.data_folder.split('/')[-3]+'_'+str(self.iteration)
            initalcommand = 'mkdir '+ self.target_dir
            p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)
        else:
            self.target_dir = target_dir
        self.model_name = self.data_folder.split('/')[-3]+'_'+str(self.iteration)+'.pth'
    def start_analyse(self,progressionmarker_background_sampling):
        print('calculate hierarchical markers.....')
        hcmarkers= get_dataset_hcmarkers(self.adata,stage_key='stage',cluster_key='leiden',use_rep='umaps')
        print('hierarchical static markers done')
        self.adata = calculateDataPathwayOverlapGene(self.adata)
        print('calculateDataPathwayOverlapGene done')
        self.adata = calculateTopPathwayGeneRanking(self.adata)
        print('calculateTopPathwayGeneRanking done')
        if not os.path.exists(os.path.join(self.target_dir,'idrem')):
            initalcommand = 'cp -r ' + os.path.join(os.path.dirname(self.data_folder),'idremResults') +' '+self.target_dir+'/idrem'
            p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)
        initalcommand = 'mkdir '+self.target_dir+'/model_save'+'&& cp ' + os.path.join(os.path.dirname(os.path.dirname(self.data_folder)),'model_save',self.model_name)+' '+self.target_dir+'/model_save/'+self.model_name
        p = subprocess.Popen(initalcommand, stdout=subprocess.PIPE, shell=True)
        if self.customized_drug is not None:
            self.adata = find_overlap_and_assign_direction(self.adata, customized_drug=self.customized_drug,cmap_dir=self.cmap_dir)
        else:
            self.adata = find_overlap_and_assign_direction(self.adata,cmap_dir=self.cmap_dir)
        if os.path.exists(os.path.join(self.target_dir,str(progressionmarker_background_sampling)+'progressionmarker_background.npy')):
            progressionmarker_background = np.load(os.path.join(self.target_dir,str(progressionmarker_background_sampling)+'progressionmarker_background.npy'),allow_pickle=True)
            progressionmarker_background = dict(progressionmarker_background.tolist())
        else:
            progressionmarker_background = get_progressionmarker_background(times=progressionmarker_background_sampling,adata= self.adata)
            np.save(os.path.join(self.target_dir,str(progressionmarker_background_sampling)+'progressionmarker_background.npy'),progressionmarker_background)
        self.adata.uns['progressionMarkers'] = runGetProgressionMarker_one_dist(os.path.join(os.path.dirname(self.data_folder),'idremResults'),progressionmarker_background,self.adata.shape[1],cutoff=0.05)
        print('Dynamic markers discovery.....done....')
        gc.collect()
        a = perturbation(self.adata, self.target_dir+'/model_save/'+self.model_name,self.target_dir+'/idrem')
        a.run('pathway',0.5,inplace=True,CUDA=True)
        a.run('drug',0.5,inplace=True)
        a.run('random_background',0.5,inplace=True)
        a.run('online_random_background',0.5,inplace=True)
        a.analysis('pathway',0.5)
        a.analysis('drug',0.5)
        a.adata.uns['hcmarkers'] = hcmarkers #get_dataset_hcmarkers(self.adata,stage_key='stage',cluster_key='leiden',use_rep='umaps')
        with open(os.path.join(self.target_dir,'attribute.pkl'),'wb') as f:
            pickle.dump(a.adata.uns,f)
        del a.adata.uns
        a.adata.obs['leiden'] = a.adata.obs['leiden'].astype(str)
        a.adata.obs['stage'] = a.adata.obs['stage'].astype(str)
        a.adata.obs['ident'] = a.adata.obs['ident'].astype(str)
        a.adata.write(self.target_dir+ '/dataset.h5ad',compression='gzip', compression_opts=9)


import multiprocessing
import numpy as np
import gc
import anndata
import pandas as pd
import scanpy as sc
from sklearn.metrics import silhouette_score, davies_bouldin_score
from sklearn.neighbors import KernelDensity
from sklearn import cluster
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from scipy.stats import norm, entropy, multivariate_normal, gamma
from scipy import stats 
import torch
from torch.nn import functional as F
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from torch.distributions.gamma import Gamma
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from scipy.stats import entropy 
from scipy import stats
from scipy.sparse import csr_matrix
from sklearn.decomposition import TruncatedSVD
import threading
from scipy.stats import multivariate_normal
import gc
def runGetSimilarity(adata):
    ids = []
    for i in range(4):

        ids.append(adata.obs[adata.obs['stage'] == i].index.tolist())
    out = [[] for _ in range(4)]
    for i in range(4):
        temp = adata[ids[i]]
        #temp2 = adata[ids[i+1]]
        out[i]= getN2CSimilarity(temp, i)

    return out
def getN2CSimilarity(adata,stagei):
    '''
    calculate top differential gene similarity between two clusters in a stage
    '''
    
    clusterIds = set(adata.obs['leiden'].values)
    
    # print(clusterIds)
    #maxj = adataj.obs['leiden'].max()
    similarities = {}
    for i,clusterIdi in enumerate(clusterIds):
        similarities[str(clusterIdi)] = []
        for j,clusterIdj in enumerate(clusterIds):
            similarities[str(clusterIdi)].append(getSimilarity(adata.uns['topGene'][str(stagei)],adata.uns['topGene'][str(stagei)],i,j))

    return similarities
def getSimilarity(adatai,adataj, i, j):
    '''
    calculate the differential gene similarity between two clusters. Differentail gene similarity = (1-Jaccard index)* gene ranking difference factor
    args: 
    adatai: anndata of stage i
    adataj: anndata of stage j
    i: the cluster id in stage i
    j: the cluster id in stage j

    return: 
    distance: the distance of top 100 differential gene between two clusters

    '''
    #iDE = adatai.d1.raw.var['features'][adatai.d1.uns['rank_genes_groups']['names'][str(i)]].values
    #jDE = adataj.d1.raw.var['features'][adataj.d1.uns['rank_genes_groups']['names'][str(j)]].values
    if isinstance(adatai[i], list):
        
        iDE = adatai[i]
    else:
        iDE = adatai[i].tolist()
    if isinstance(adataj[j], list):
        
        jDE = adataj[j]
    else:
        jDE = adataj[j].tolist()
  

    intersection = 0
    orderdistance = 0
    for i, each in enumerate(iDE):
        if each in jDE:
            
            intersection+=1
            jindex = jDE.index(each)
        
            orderdistance += abs(jindex-i)
        else:
            orderdistance+=100
    union = len(iDE)+len(jDE) - intersection
    JaccardIndex = intersection/union
    # print(intersection,union)
   
    orderdistance = orderdistance/(100*len(jDE))
    # print(orderdistance)

    return (1-JaccardIndex)*orderdistance
def normalizeDistanceRE(distance):#good
    '''
    Normalize the kl divergence distance and top differential gene distances. (Use the z-score method.)
    
    args: 
    distance: a list of distance metrics. ([gaussian kl divergence, top differential gene difference])
    
    return:
    normalizedDistance: sum of normalized kl divergence and top differential gene difference distance
    '''
    
    distance = np.array(distance)
    
    gaussiankld = np.array(distance[:,0].tolist()).reshape(-1,)
    
    #gammakld = np.array(distance[:,1].tolist()).reshape(-1,)

    de = np.array(distance[:,1].tolist())

    
    gaussian_kl_zscore = gaussiankld#stats.zscore(gaussiankld)
    #gamma_kl_zscore = stats.zscore(gammakld)
    de_zscore = stats.zscore(de)
    #ggkld = np.array(gaussiankld)+np.array(gammakld)
    #kld_zscore = stats.zscore(ggkld)
    return gaussian_kl_zscore+de_zscore#gaussian_kl_zscore+gamma_kl_zscore+de_zscore

def calculateN2CKLDistance(mu,sigma,previousRep,topGeneD):
    '''
    calculate the distance between a cell and all clusters in the same stage
    args: mu, sigma is the attribute of one cell
    previousRep: previous representation of clusters in one stage [#cluster,[gaussian(3 parameters)*#hidden]
    topGeneD: top gene distances between two clusters of one stage
    '''
    mean1 = mu
    std1 = sigma
    #mean1 = cluster1[:,0]
    covariance1 = torch.tensor(np.diag(std1**2))
    p = MultivariateNormal(torch.tensor(mean1), covariance1)
    d = []

    for no, each in enumerate(previousRep):
        
        cluster2 = np.array(each[0])
        mean2 = cluster2[:,0]
        std2 = cluster2[:,1]
        covariance2 = torch.tensor(np.diag(std2**2))
        q = MultivariateNormal(torch.tensor(mean2), covariance2)
        kl = torch.distributions.kl.kl_divergence(p, q)



        d.append([kl.detach().numpy(),topGeneD[no]])
    out = normalizeDistanceRE(d)
    return out

class calculateN2CKLDProcess(multiprocessing.Process):
    def __init__(self, mu,sigma,previousRep,topGeneD,ddd,tempids,idnumber,currentCluster,out,cellid):
        multiprocessing.Process.__init__(self)
        self.mu = mu
        self.sigma = sigma
        self.previousRep = previousRep
        self.topGeneD = topGeneD
        self.ddd = ddd
        self.tempids = tempids
        self.idnumber = idnumber
        self.currentCluster = currentCluster
        self.out = out
    def run(self):
        #time.sleep(10)
        time1 = time.time()
        temp = calculateN2CKLDistance(self.mu,self.sigma,self.previousRep,self.topGeneD)
        time2 = time.time()
        # print(time2-time1)
        idx = np.argmin(np.array(temp))
        self.out[cellid] = [idx, self.idnumber]
        #self.out[]
        #if idx != int(self.currentCluster):
        #    self.ddd.append(idx)
        #    self.tempids.append(self.idnumber)
def mcSampling(mus, sigmas):
    '''
    monte carlo sampling strategy to sample data points from a gaussian distribution in each cell,
    for example the hidden space is 10, then sample 100 data point from input cell. If the number of cell is 200, then
    the sampled matrix will be [10, 200*100]
    args:
    mus, sigmas: mu and sigma vectors of input cells (shape: [number of cell, number of hidden nodes])
    
    return:
    samplegaussian: a list of sampled gaussian datapoints
    '''
    samplegaussian= [[] for _ in range(64)]
    hiddensize=64#len(mus[0])
    mus = np.vstack(mus)
    sigmas=np.vstack(sigmas)

    for i in range(hiddensize):
        
        mean = mus[:,i]
        mean = mean.repeat(100)
        std = sigmas[:,i]
        std = std.repeat(100)
        normal = Normal(torch.zeros(size=mean.shape)+torch.tensor(mean),torch.zeros(size=std.shape)+torch.tensor(std)).sample()
        samplegaussian[i].append(normal.numpy())
        '''for j in range(len(mus[i])):
            
            mean = mus[i][j]
            theta = thetas[i][j]
            std = np.exp(sigmas[i][j] / 2)
            #print(theta)
            normal = Normal(torch.zeros((2))+torch.tensor(mean),torch.zeros((2))+torch.tensor(std)).sample()
            gamma = Gamma(torch.ones(2), torch.zeros(2)+1/theta).sample()
#minimal unit test 100->2
            samplegamma[j].extend(gamma.numpy())
            samplegaussian[j].extend(normal.numpy())'''

    return samplegaussian

class GaussianRepThread(threading.Thread):
    def __init__(self, output, data,i):
        threading.Thread.__init__(self)
        self.i = i
        self.data = data
        self.output = output
    def run(self):
        #time.sleep(10)
        self.output[self.i] = norm.fit(self.data)
def fitClusterGaussianRepresentation(data):
    '''
    fit gaussian distributions for each hidden node
    args: 
    data: samples of the gaussian distribution of each hidden node
    return:
    out: list of mu and sigma of each hidden node
    '''
    threads = []
    out = [[] for _ in range(len(data))]
    for i in range(len(data)):
        threads.append(GaussianRepThread(out,data[i],i))
    for each in threads:
        each.start()
    for each in threads:
        each.join()
    return out
class GammaRepThread(threading.Thread):
    def __init__(self, output, data,i):
        threading.Thread.__init__(self)
        self.i = i
        self.data = data
        self.output = output
    def run(self):
        #time.sleep(10)
        self.output[self.i] = gamma.fit(self.data)
def fitClusterGammaRepresentation(data):
    '''
    Fit gamma distributions for each hidden node
    args: 
    data: samples of the gamma distribution of each hidden node
    return:
    out: list of alpha, beta of each hidden node
    '''
    threads = []
    out = [[] for _ in range(len(data))]
    for i in range(len(data)):

        
        threads.append(GammaRepThread(out,data[i][0],i))
    for each in threads:
        each.start()
    for each in threads:
        each.join()
    return out

def getClusterRepresentation(mus, sigmas):
    '''
    MC strategy to sample gaussian data points. Use sampled data points to fit gaussian distributions
    args: 
    mus, sigmas: mu and sigma vectors of input cells (shape: [number of cell, number of hidden nodes])
   
    return: 
    fitClusterGaussianRepresentation(sampling[0]):
    a list of fitted gaussian distributions
    '''
    T1 = time.time()
    sampling = mcSampling(mus,sigmas)
    T2 = time.time()
    #print('time is %s ' %((T2-T1)))
    return fitClusterGaussianRepresentation(sampling)
def clustertype(adata):
    '''
    find the most common cell types to represent the cluster
    args:
    adata: anndata of one cluster
    
    return: the most common cell types in the cluster
    '''
    dic = {}
    for each in adata.obs['cell.type.ident.fine']:
        if each not in dic.keys():
            dic[each]=1
        else:
            dic[each]+=1
    return max(dic, key=dic.get)
def normalizeDistance(distance):
    '''
    Normalize the kl divergence distance and top differential gene distances. (Use the z-score method.)
    
    args: 
    distance: a list of distance metrics. ([gaussian kl divergence, top differential gene difference])
    
    return:
    normalizedDistance: sum of normalized kl divergence and top differential gene difference distance
    '''
    

    distance = np.array(distance)
    
    gaussiankld = np.array(distance[:,0])

    de = np.array(distance[:,1])
    if gaussiankld.shape[0] != 1:
        # de_zscore = stats.zscore(de)
        #min max normalization
        de_minmax = (de - de.min()) / (de.max() - de.min())
        ggkld = gaussiankld
        #min max normalization
        ggkld_minmax = (ggkld - ggkld.min()) / (ggkld.max() - ggkld.min())
        # ggkld_zscore = stats.zscore(ggkld)
    else:
        ggkld_minmax= gaussiankld
        de_minmax = de
    
    
    return ggkld_minmax+de_minmax#gaussian_kl_zscore+gamma_kl_zscore+de_zscore
def calculateKL(cluster1gaussian, cluster2gaussian):
    '''
    calculate KL divergence of multivariate gaussian distributions between two clusters.
    args: 
    cluster1gaussian: a list of [mean, std] of gaussian distribution of cluster 1

    cluster2gaussian:a list of [mean, std] of gaussian distribution of cluster 2

    
    return: kl divergence of multivariate gaussian distributions
    
    '''
    cluster1 = np.array(cluster1gaussian).reshape(-1,2)
    cluster2 = np.array(cluster2gaussian).reshape(-1,2)
    
    std1 = cluster1[:,1]
    covariance1 = torch.tensor(np.diag(std1**2))

    mean1 = cluster1[:,0]

    p = MultivariateNormal(torch.tensor(mean1), covariance1)
    mean2 = cluster2[:,0]
    std2 = cluster2[:,1]
    covariance2 = torch.tensor(np.diag(std2**2))
    q = MultivariateNormal(torch.tensor(mean2), covariance2)
    kl = torch.distributions.kl.kl_divergence(p, q)
    
  

    return kl.detach().numpy()

import time






from typing import Optional, Sequence, Dict, Any
import scipy.cluster.hierarchy as sch
from .distDistance import calculateKL
import scanpy as sc
import numpy as np
from pandas.api.types import CategoricalDtype
import pandas as pd
from scanpy.tools._utils import _choose_representation
from anndata import AnnData
from .distDistance import fitClusterGaussianRepresentation

def getclusterreps(data):
    '''
    input: pandas dataframe, index is cluster id. rows are cells, columns are hidden representation 
    outs: dictionary {leiden: sampled Norm distribution (64 mu and sigmas)}
    '''
    clusterid = data.index.unique().tolist()
    out = {}
    for each in clusterid:
        out[each] = fitClusterGaussianRepresentation(data.loc[each].T.values)
    return out
def calculateDistance(reps):
    outs = {}
    for i, each1 in enumerate(list(reps.keys())):
        outs[each1] = []
        for j, each2 in enumerate(list(reps.keys())):
            outs[each1].append(((calculateKL(reps[each1], reps[each2])+calculateKL(reps[each2], reps[each1]))/2))
    df = pd.DataFrame(outs,index=list(reps.keys()))
    df.columns = list(reps.keys())
    df.index = df.index.astype(int)
    df = df.sort_index()
    df.columns =  df.columns.astype(int)
    df = df.sort_index(axis=1)
    return df
def calculateDistanceUMAP(umaps):
    outs = {}

    for i, each1 in enumerate(list(umaps.index)):
        outs[each1] = []
        
        for j, each2 in enumerate(list(umaps.index)):
            outs[each1].append(np.linalg.norm(umaps.loc[each1]- umaps.loc[each2]))
    df = pd.DataFrame(outs,index=list(umaps.index))
    df.columns = list(umaps.index)
    df.index = df.index.astype(int)
    df = df.sort_index()
    df.columns =  df.columns.astype(int)
    df = df.sort_index(axis=1)
    return df
def mydendrogram(
    adata: AnnData,
    groupby: str,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    var_names: Optional[Sequence[str]] = None,
    use_raw: Optional[bool] = None,
    cor_method: str = "pearson",
    linkage_method: str = "complete",
    optimal_ordering: bool = False,
    key_added: Optional[str] = None,
    inplace: bool = True,
) -> Optional[Dict[str, Any]]:
    """\
    Computes a hierarchical clustering for the given `groupby` categories.

    By default, the PCA representation is used unless `.X`
    has less than 50 variables.

    Alternatively, a list of `var_names` (e.g. genes) can be given.

    Average values of either `var_names` or components are used
    to compute a correlation matrix.

    The hierarchical clustering can be visualized using
    :func:`scanpy.pl.dendrogram` or multiple other visualizations that can
    include a dendrogram: :func:`~scanpy.pl.matrixplot`,
    :func:`~scanpy.pl.heatmap`, :func:`~scanpy.pl.dotplot`,
    and :func:`~scanpy.pl.stacked_violin`.

    .. note::
        The computation of the hierarchical clustering is based on predefined
        groups and not per cell. The correlation matrix is computed using by
        default pearson but other methods are available.

    Parameters
    ----------
    adata
        Annotated data matrix
    {n_pcs}
    {use_rep}
    var_names
        List of var_names to use for computing the hierarchical clustering.
        If `var_names` is given, then `use_rep` and `n_pcs` is ignored.
    use_raw
        Only when `var_names` is not None.
        Use `raw` attribute of `adata` if present.
    cor_method
        correlation method to use.
        Options are 'pearson', 'kendall', and 'spearman'
    linkage_method
        linkage method to use. See :func:`scipy.cluster.hierarchy.linkage`
        for more information.
    optimal_ordering
        Same as the optimal_ordering argument of :func:`scipy.cluster.hierarchy.linkage`
        which reorders the linkage matrix so that the distance between successive
        leaves is minimal.
    key_added
        By default, the dendrogram information is added to
        `.uns[f'dendrogram_{{groupby}}']`.
        Notice that the `groupby` information is added to the dendrogram.
    inplace
        If `True`, adds dendrogram information to `adata.uns[key_added]`,
        else this function returns the information.

    Returns
    -------
    If `inplace=False`, returns dendrogram information,
    else `adata.uns[key_added]` is updated with it.

    Examples
    --------
    >>> import scanpy as sc
    >>> adata = sc.datasets.pbmc68k_reduced()
    >>> sc.tl.dendrogram(adata, groupby='bulk_labels')
    >>> sc.pl.dendrogram(adata, groupby='bulk_labels')
    <Axes: >
    >>> markers = ['C1QA', 'PSAP', 'CD79A', 'CD79B', 'CST3', 'LYZ']
    >>> sc.pl.dotplot(adata, markers, groupby='bulk_labels', dendrogram=True)
    """
    if isinstance(groupby, str):
        # if not a list, turn into a list
        groupby = [groupby]
    for group in groupby:
        if group not in adata.obs_keys():
            raise ValueError(
                "groupby has to be a valid observation. "
                f"Given value: {group}, valid observations: {adata.obs_keys()}"
            )
        if not isinstance(adata.obs[group].dtype, CategoricalDtype):
            raise ValueError(
                "groupby has to be a categorical observation. "
                f"Given value: {group}, Column type: {adata.obs[group].dtype}"
            )

    if var_names is None:
        rep_df = pd.DataFrame(
            _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs)
        )
        categorical = adata.obs[groupby[0]]
        if len(groupby) > 1:
            for group in groupby[1:]:
                # create new category by merging the given groupby categories
                categorical = (
                    categorical.astype(str) + "_" + adata.obs[group].astype(str)
                ).astype("category")
        categorical.name = "_".join(groupby)

        rep_df.set_index(categorical, inplace=True)
        categories = rep_df.index.categories
    else:
        gene_names = adata.raw.var_names if use_raw else adata.var_names

    # aggregate values within categories using 'mean'
    
    mean_df = rep_df.groupby(level=0).mean()

    import scipy.cluster.hierarchy as sch
    from scipy.spatial import distance
    test = getclusterreps(rep_df)
    test2 = calculateDistance(test)
    corr_matrix = mean_df.T.corr(method=cor_method)
    corr_condensed = distance.squareform(test2)

    z_var = sch.linkage(
        corr_condensed, method=linkage_method, optimal_ordering=optimal_ordering
    )
    dendro_info = sch.dendrogram(z_var, labels=list(categories), no_plot=True)

    dat = dict(
        linkage=z_var,
        groupby=groupby,
        use_rep=use_rep,
        cor_method=cor_method,
        linkage_method=linkage_method,
        categories_ordered=dendro_info["ivl"],
        categories_idx_ordered=dendro_info["leaves"],
        dendrogram_info=dendro_info,
        correlation_matrix=corr_matrix.values,
    )

    if inplace:
        if key_added is None:
            key_added = f'dendrogram_{"_".join(groupby)}'
        # logg.info(f"Storing dendrogram info using `.uns[{key_added!r}]`")
        adata.uns[key_added] = dat
    else:
        return dat
def mydendrogramUMAP(
    adata: AnnData,
    groupby: str,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    var_names: Optional[Sequence[str]] = None,
    use_raw: Optional[bool] = None,
    cor_method: str = "pearson",
    linkage_method: str = "complete",
    optimal_ordering: bool = False,
    key_added: Optional[str] = None,
    inplace: bool = True,
) -> Optional[Dict[str, Any]]:
    """\
    Computes a hierarchical clustering for the given `groupby` categories.

    By default, the PCA representation is used unless `.X`
    has less than 50 variables.

    Alternatively, a list of `var_names` (e.g. genes) can be given.

    Average values of either `var_names` or components are used
    to compute a correlation matrix.

    The hierarchical clustering can be visualized using
    :func:`scanpy.pl.dendrogram` or multiple other visualizations that can
    include a dendrogram: :func:`~scanpy.pl.matrixplot`,
    :func:`~scanpy.pl.heatmap`, :func:`~scanpy.pl.dotplot`,
    and :func:`~scanpy.pl.stacked_violin`.

    .. note::
        The computation of the hierarchical clustering is based on predefined
        groups and not per cell. The correlation matrix is computed using by
        default pearson but other methods are available.

    Parameters
    ----------
    adata
        Annotated data matrix
    {n_pcs}
    {use_rep}
    var_names
        List of var_names to use for computing the hierarchical clustering.
        If `var_names` is given, then `use_rep` and `n_pcs` is ignored.
    use_raw
        Only when `var_names` is not None.
        Use `raw` attribute of `adata` if present.
    cor_method
        correlation method to use.
        Options are 'pearson', 'kendall', and 'spearman'
    linkage_method
        linkage method to use. See :func:`scipy.cluster.hierarchy.linkage`
        for more information.
    optimal_ordering
        Same as the optimal_ordering argument of :func:`scipy.cluster.hierarchy.linkage`
        which reorders the linkage matrix so that the distance between successive
        leaves is minimal.
    key_added
        By default, the dendrogram information is added to
        `.uns[f'dendrogram_{{groupby}}']`.
        Notice that the `groupby` information is added to the dendrogram.
    inplace
        If `True`, adds dendrogram information to `adata.uns[key_added]`,
        else this function returns the information.

    Returns
    -------
    If `inplace=False`, returns dendrogram information,
    else `adata.uns[key_added]` is updated with it.

    Examples
    --------
    >>> import scanpy as sc
    >>> adata = sc.datasets.pbmc68k_reduced()
    >>> sc.tl.dendrogram(adata, groupby='bulk_labels')
    >>> sc.pl.dendrogram(adata, groupby='bulk_labels')
    <Axes: >
    >>> markers = ['C1QA', 'PSAP', 'CD79A', 'CD79B', 'CST3', 'LYZ']
    >>> sc.pl.dotplot(adata, markers, groupby='bulk_labels', dendrogram=True)
    """
    if isinstance(groupby, str):
        # if not a list, turn into a list
        groupby = [groupby]
    for group in groupby:
        if group not in adata.obs_keys():
            raise ValueError(
                "groupby has to be a valid observation. "
                f"Given value: {group}, valid observations: {adata.obs_keys()}"
            )
        if not isinstance(adata.obs[group].dtype, CategoricalDtype):
            raise ValueError(
                "groupby has to be a categorical observation. "
                f"Given value: {group}, Column type: {adata.obs[group].dtype}"
            )

    if var_names is None:
        rep_df = pd.DataFrame(
            _choose_representation(adata, use_rep=use_rep, n_pcs=n_pcs)
        )
        categorical = adata.obs[groupby[0]]
        if len(groupby) > 1:
            for group in groupby[1:]:
                # create new category by merging the given groupby categories
                categorical = (
                    categorical.astype(str) + "_" + adata.obs[group].astype(str)
                ).astype("category")
        categorical.name = "_".join(groupby)

        rep_df.set_index(categorical, inplace=True)
        categories = rep_df.index.categories
    else:
        gene_names = adata.raw.var_names if use_raw else adata.var_names

    # aggregate values within categories using 'mean'
    
    mean_df = rep_df.groupby(level=0).mean()

    import scipy.cluster.hierarchy as sch
    from scipy.spatial import distance
    # test = getclusterrepsUMAP(rep_df)

    test2 = calculateDistanceUMAP(mean_df)
    # corr_matrix = mean_df.T.corr(method=cor_method)
    corr_condensed = distance.squareform(test2)

    z_var = sch.linkage(
        corr_condensed, method=linkage_method, optimal_ordering=optimal_ordering
    )
    dendro_info = sch.dendrogram(z_var, labels=list(categories), no_plot=True)

    dat = dict(
        linkage=z_var,
        groupby=groupby,
        use_rep=use_rep,
        cor_method=cor_method,
        linkage_method=linkage_method,
        categories_ordered=dendro_info["ivl"],
        categories_idx_ordered=dendro_info["leaves"],
        dendrogram_info=dendro_info,
        # correlation_matrix=corr_matrix.values,
    )

    if inplace:
        if key_added is None:
            key_added = f'dendrogram_{"_".join(groupby)}'
        # logg.info(f"Storing dendrogram info using `.uns[{key_added!r}]`")
        adata.uns[key_added] = dat
    else:
        return dat
def get_siblings_at_each_level(leaf_id, Z):
    n = Z.shape[0] + 1  # The total number of initial leaves
    clusters = {i: [i] for i in range(n)}  # Initialize clusters with single leaves
    siblings = []

    # Helper function to find the parent node for the given leaf or cluster
    def find_parent(node):
        for i, (left, right, _, _) in enumerate(Z):
            if node in [left, right]:
                return i + n, left if node != left else right
        return None, None

    # Build clusters and collect siblings at different levels
    for k in range(n - 1):
        left, right = int(Z[k, 0]), int(Z[k, 1])
        clusters[n + k] = clusters[left] + clusters[right]

    # Trace back from the leaf to the root and collect siblings
    current_node = leaf_id
    while True:
        parent, sibling_node = find_parent(current_node)
        if parent is None:  # Reached the root
            break
        # Check if sibling_node is a leaf or a cluster and collect all leaves
        sibling_leaves = clusters[sibling_node]
        siblings.append(sibling_leaves)
        current_node = parent

    return siblings[::-1]  # Reverse to have siblings from bottom to top
def map_leaves_to_label(categories_idx_ordered, categories_ordered):
    table = {}
    for i, each in enumerate(categories_idx_ordered):
        table[each] = int(categories_ordered[i])
    return table
def my_logfold_change(adata1, adata2,log_fold_change_cutoff=None,abs = False):
    genenames = adata1.var_names.tolist()
    mean1 = np.mean(adata1.X.toarray(),axis=0)
    mean2 = np.mean(adata2.X.toarray(),axis=0)
    logfold_change = mean2 - mean1
    logfold_change_dict = {}
    order = np.argsort(logfold_change)[::-1]
    mean2 = mean2[order]
    mean1 = mean1[order]
    for i, each in enumerate(genenames):
        logfold_change_dict[each] = logfold_change[i]
    if abs ==True:
        temp = {k: v for k, v in sorted(logfold_change_dict.items(), key=lambda item: np.abs(item[1]),reverse=True)}
    else:
        temp = {k: v for k, v in sorted(logfold_change_dict.items(), key=lambda item: item[1],reverse=True)}

    df = pd.DataFrame()
    df['log_fold_changes'] = temp.values()
    df['names'] = temp.keys()

    if log_fold_change_cutoff is not None:
        df = df[np.abs(df['log_fold_changes'])>log_fold_change_cutoff]
    df['Mean_compared_clusters'] = mean1
    df['Mean_selected_clusters'] = mean2
    return df
def build_reference_siblings(dendrogram):
    total_leaves = len(dendrogram['dendrogram_info']['ivl'])
    table = map_leaves_to_label(dendrogram['categories_idx_ordered'], dendrogram['categories_ordered'])
    out = {}
    for leave_id in range(total_leaves):
        siblings = get_siblings_at_each_level(leave_id, dendrogram['linkage'])
        leaf_node_id = table[leave_id]
        out[str(leaf_node_id)] = {}
        level_table = {}
        memory = []
        for i, sibs in enumerate(siblings[::-1]):
            translated_sibs = [table[each] for each in sibs]
            memory+=translated_sibs
            level_table[len(siblings)-1-i] = memory.copy()
        for level, sibs in enumerate(siblings):
            out[str(leaf_node_id)][str(level)] = level_table[level]
    return out
def gethcmarkers(adata, selected,groups):
    ref = [[selected],[selected]+groups]
    ids = []
    ids = adata[adata.obs['leiden'] == str(selected)].obs.index.tolist()
    selectedids = ids.copy()
    groupids = []
    for each in groups:
        groupids +=adata[adata.obs['leiden'] == str(each)].obs.index.tolist()
    ids += groupids.copy()
    new = adata[ids]
    new.obs['compare'] = None
    new.obs.loc[new.obs.index.isin(selectedids), "compare"] = 'selected'
    new.obs.loc[new.obs.index.isin(groupids), "compare"] = 'background'
    sc.tl.rank_genes_groups(new,groupby='compare',method='wilcoxon',rankby_abs=False)
    selected = new[selectedids]
    background = new[groupids]
    df1 = my_logfold_change(background,selected)
    df1 = df1.set_index('names')
    df = pd.DataFrame.from_dict({key: new.uns['rank_genes_groups'][key]['selected'] for key in['names','pvals_adj','scores']})
    df = df.set_index('names')
    df = df.join(df1)
    positive_marker = df[df['log_fold_changes']>0]
    negative_marker = df[df['log_fold_changes']>0]
    positive_marker= positive_marker[positive_marker['pvals_adj'] <0.05]
    negative_marker= negative_marker[negative_marker['pvals_adj'] <0.05]
    
    return positive_marker, negative_marker,ref
def get_stage_hcmarkers(adata,key,use_rep,stage):
    
    if use_rep == 'z':
        mydendrogram(adata,use_rep='z',groupby=key)
    elif use_rep == 'umaps':
        mydendrogramUMAP(adata,use_rep='X_umap',groupby=key)
    
    test = build_reference_siblings(adata.uns['dendrogram_'+key])
    Z = adata.uns['dendrogram_'+key]['linkage']
    order = adata.uns['dendrogram_'+key]['categories_ordered']
    out_table = {}
    for each in list(test.keys()):
        
        out_table[each] = {}
        for level in test[each]:
            out_table[each][level] = {}
            out_table[each][level]['chosen'] = {}
            pos_markers, neg_markers, ref = gethcmarkers(adata,int(each),test[each][level])
            out_table[each][level]['chosen']['pos'] = pos_markers
            out_table[each][level]['chosen']['neg'] = neg_markers
            out_table[each][level]['ref'] = ref
    return Z,order,out_table
def get_dataset_hcmarkers(adata,stage_key,cluster_key,use_rep):
    hcmarkers = {}
    for each in list(adata.obs[stage_key].unique()):
        hcmarkers[str(each)] = {}
        stage_adata = adata[adata.obs[stage_key] == each]
        Z,order,markers = get_stage_hcmarkers(stage_adata,cluster_key,use_rep=use_rep,stage=each)
        hcmarkers[str(each)]['Z'] = Z
        hcmarkers[str(each)]['order'] = order
        hcmarkers[str(each)]['markers'] = markers
    return hcmarkers
# get_dataset_hcmarkers(adata,'stage','leiden')
#unit test
if __name__ == '__main__':
    import pickle
    # adata = sc.read_h5ad('/mnt/md0/yumin/UNAGI_old/data/mes/4/org_test/dataset.h5ad')
    adata = sc.read('/mnt/md0/yumin/UNAGI_pyro/src/small_0/dataset.h5ad')
    # adata.uns = pickle.load(open('/mnt/md0/yumin/UNAGI_old/data/mes/4/org_test/attribute.pkl', 'rb'))
    adata.uns = pickle.load(open('/mnt/md0/yumin/UNAGI_pyro/src/small_0/attribute.pkl', 'rb'))
    
    adata.uns['hcmarkers'] = get_dataset_hcmarkers(adata,stage_key='stage',cluster_key='leiden',use_rep='umaps')
    # with open('/mnt/md0/yumin/UNAGI_old/data/mes/4/org_test/attribute_1109_newhc.pkl','wb') as f:
    #     pickle.dump(adata.uns,f)


import pandas as pd
def get_top_compounds(adata, intensity, top_n=None, cutoff=None):
    if top_n is not None:
        return pd.DataFrame.from_dict(adata.uns['drug_perturbation_score'][str(intensity)]['total']['top_compounds'])[:top_n]
    elif cutoff is not None:
        temp = pd.DataFrame.from_dict(adata.uns['drug_perturbation_score'][str(intensity)]['total']['top_compounds'])
        return temp[temp['pval_adjusted'] < cutoff]
    else:
        print('Please specify top_n or cutoff')


import scanpy as sc
import numpy as np
import umap
import pandas as pd
import os
import pickle
from scipy.stats import rankdata
import torch
from .distDistance import getClusterRepresentation
from sklearn.neighbors import kneighbors_graph
def get_data_file_path(filename):
    file_path = os.path.join(os.path.dirname(__file__), 'data', filename)
    return file_path
def split_dataset_into_stage(adata_path, folder, key):
    '''
    split dataset into stages and write to the path
    args:
    adata: IPF database
    key: key of the dataset
    path: path to write the dataset
    
    return:
    None
    '''
    adata = sc.read_h5ad(adata_path)
    for each in list(adata.obs[key].unqiue()):
        adata_temp = adata[adata.obs[key] == each]
        adata_temp.write_h5ad(os.path.join(folder,'%s.h5ad'%each),compression='gzip',compression_opts=9)
def transfer_to_ranking_score(gw):
    '''
    ranking score
    '''
    # gw = adata.layers['geneWeight'].toarray()
    od = gw.shape[1]-rankdata(gw,axis=1)+1
    score = 1+1/np.power(od,0.5)
    
    return score
def clustertype_old(adata):
    '''
    find the most common cell types to represent the cluster
    args:
    adata: anndata of one cluster
    
    return: the most common cell types in the cluster
    '''
    dic = {}
    for each in adata.obs['ident']:
        if each not in dic.keys():
            dic[each]=1
        else:
            dic[each]+=1
    #print(dic)        
    return max(dic, key=dic.get)
def clustertype40(adata):
    '''
    annotate the cluster with cells >40% if no one >40%, annotate with the highest one
    args:
    adata: anndata of one cluster
    
    return: the most common cell types in the cluster
    '''
    dic = {}
    total = 0
    
    for each in adata.obs['name.simple']:
        if each not in dic.keys():
            dic[each]=1
        else:
            dic[each]+=1
        total+=1
    #print(dic)
   
    anootate = ''
    
    flag = False #flag to see if there are more than 1 cell types > 40%
    for each in list(dic.keys()):
        
        if dic[each] > total*0.5:
            if flag == False:
                anootate+=each
                flag = True
            else:
                anootate+='/'+each
    if flag == False:
        for each in list(dic.keys()):
        
            if dic[each] > total*0.4:
                if flag == False:
                    anootate+=each
                    flag = True
                else:
                    anootate+='/'+each
        if flag == False:
            for each in list(dic.keys()):
        
                if dic[each] > total*0.3:
                    if flag == False:
                        anootate+=each
                        flag = True
                    else:
                        anootate+='/'+each
        if flag == False:
            for each in list(dic.keys()):
        
                if dic[each] > total*0.2:
                    if flag == False:
                        anootate+=each
                        flag = True
                    else:
                        anootate+='/'+each
        if flag == False:
            anootate = 'Mixed'
    return anootate

def changeCluster(adata, ids, newIds):
    '''
    re-assign cluster id to cells
    args: 
    adata: IPF database
    ids: cell id in one stage needed to be changed 
    vamos: new cluster id of cells
    return: 
    out: new IPF database
    '''

    for i, cluster in enumerate(ids):
        for j, each in enumerate(cluster):
            #print('vamos[i][j]',vamos[i][j])
            adata.obs['leiden'][each] = newIds[i][j]
    return adata
def extracth5adcluster(data, ID):
    '''
    extract data from a certain cluster
    args:
    data: data (H5AD class) of a certain stage
    ID: the id of cluster
    
    return:
    splitData: data of a certain cluster
    '''
    splitDataID = []
    splitData = []
    for i in range(len(data)):
        
        if int(data.obs['leiden'][i]) == ID:
            splitDataID.append(i)
    splitData = data[splitDataID]
    
    return splitData,splitDataID

def extracth5adclusterids(data, ID):
    '''
    extract index from a certain cluster
    args:
    data: data (H5AD class) of a certain stage
    ID: the id of cluster
    
    return:
    splitData: data of a certain cluster
    '''
    splitDataID = []
    splitData = []
    for i in range(len(data)):
        
        if int(data.obs['leiden'][i]) == ID:
            splitDataID+=data[i].obs.index.tolist()
    splitData = data[splitDataID]
    
    return splitDataID

def retrieveResults(adata, ids,stage):
    '''
    retrieve results from adata
    args:
    adata: IPF database
    ids: id of cells in each stage
    stage: stage id
    out:
    Rep
    cluster_type
    top_gene
    average_value
    '''
    Rep = []
    average_value = []
    cluster_type = []
    stageids = adata.obs[adata.obs['stage'] == stage].index.tolist()
    stageadata = adata[stageids]
    stageadata = sc.tl.rank_genes_groups(stageadata, 'leiden', method='wilcoxon',n_genes=100,copy=True)

    top_gene = []
    idskeys = list(ids.keys())
    clusteridset = set(stageadata.obs['leiden'].values)


    for i in ids.keys():
        if int(i) not in clusteridset:
            print('miss1')
            print(i)
            #ids.pop(i, None)
            continue
        tempid = stageadata.obs[stageadata.obs['leiden'] == int(i)].index.tolist()
        if len(tempid) == 0:
            #ids.pop(i, None)
            print('miss2')
            print(i)
            continue
        cluster = stageadata[tempid]
        Rep.append(getClusterRepresentation(cluster.obs['mu'].values,cluster.obs['sigma'].values,cluster.obs['theta'].values))
        average_value.append(np.mean(cluster.X,axis =0))
        cluster_type.append(clustertype(cluster))
        top_gene.append(list(stageadata.uns['rank_genes_groups']['names'][str(i)]))
    print('retrive ok')
    return Rep, cluster_type, top_gene, average_value
def getAvgClusgerRep(mid, iteration):
    avg = []
    for i in range(4):
        tempavg = []
        adata = sc.read_h5ad('./'+mid+'/'+str(iteration)+'/stagedata/%d.h5ad'%(i))
        for clusterid in set(adata.obs['leiden']):
            clusteradataid = adata.obs[adata.obs['leiden'] == clusterid].index.tolist()
            clusteradata = adata[clusteradataid]
            clusteravg = np.mean(clusteradata.X, axis = 0)
            clusteravg = np.reshape(np.array(clusteravg),-1)
            tempavg.append(clusteravg)
        tempavg = np.array(tempavg)
        avg.append(tempavg)
    avg = np.array(avg)
    return avg
def updateAttributes(adata,reps):
    '''
    update stage, cluster id of each stage, top 100 differential genes, cell types of clusters to anndata

    args: 
    adata: anndata of database
    stage: the IPF stage of this fold of data
    
    
    returns: 
    adata: updated anndata of database
    rep: representations of sampled gaussian data points
    average_cluster: cluster.X average values
    '''
    #stageids = adata.obs[adata.obs['stage'] == stage].index.tolist()
    #stageadata = adata[stageids]
    print('top gene')
    sc.tl.rank_genes_groups(adata, 'leiden', method='wilcoxon',n_genes=100)
    print('done')
    clusters = set(adata.obs['leiden'])
    average_cluster= []
    TDG = []
    rep = []
    celltypes = []
    
    # adata.obs['name.simple'] = adata.obs['name.simple'].astype('string')
    for i in range(len(clusters)):
        
        clusteradata,ids = extracth5adcluster(adata,i)
        clutertype = clustertype40(clusteradata) #use >40 strategy
        celltypes.append(clutertype)
        one_mean = np.mean(clusteradata.X,axis =0)
        one_mean = np.reshape(np.array(one_mean),-1)
        average_cluster.append(one_mean)
        adata.obs['ident'] = adata.obs['ident'].astype(str)
        TDG.append(list(adata.uns['rank_genes_groups']['names'][str(i)]))

        rep.append(getClusterRepresentation(reps[0][ids],reps[1][ids]))
        ids = extracth5adclusterids(adata,i)
        adata.obs.loc[adata.obs.index.isin(ids), "ident"] = clutertype
        # print('cluster type showing')
        # print(clutertype)
        # print(adata[ids].obs['ident'])
    adata.obs['ident'] = adata.obs['ident'].astype(str)
    adata.uns['clusterType'] = celltypes
    adata.uns['topGene'] = TDG
    return adata,average_cluster,rep
def saveRep(rep):
    '''
    write rep of all stages to disk
    '''
    np.save('./stagedata/rep.npy',np.array(rep,dtype=object))
def saveRep(rep,midpath,iteration):
    '''
    write rep of all stages to disk with midpath in iterative training
    '''
    np.save(os.path.join(midpath,str(iteration)+'/stagedata/rep.npy'),np.array(rep,dtype=object))

import scipy.sparse as sp
from scipy.sparse import csr_matrix
import anndata
import os

def get_all_adj_adata(adatas):
    data_size = []
    obs = []
    for i, each in enumerate(adatas):
        each.obs['stage'] = i
        obs.append(each.obs)
        each.obsp['gcn_connectivities'] = each.obsp['gcn_connectivities'].tocoo()
        if i == 0:
            if sp.isspmatrix(each.X):
                X = each.X
            else:
                X = csr_matrix(each.X)
            data_size.append(each.obsp['gcn_connectivities'].shape[0])
            gcn_data = each.obsp['gcn_connectivities'].data.tolist()
            col = each.obsp['gcn_connectivities'].col.tolist()
            row = each.obsp['gcn_connectivities'].row.tolist()
            if 'geneWeight' in each.layers.keys():
                geneWeight = each.layers['geneWeight']
            if 'X_pca' in each.obsm.keys():
                pca = each.obsm['X_pca']

        else:
            X = sp.vstack((X, each.X), format='csr')
            
            gcn_data += each.obsp['gcn_connectivities'].data.tolist()
            col += (data_size[i-1] + each.obsp['gcn_connectivities'].col).tolist()
            row += (data_size[i-1] + each.obsp['gcn_connectivities'].row).tolist()
            data_size.append(each.obsp['gcn_connectivities'].shape[0] + data_size[i - 1])
            if 'geneWeight' in each.layers.keys():
                geneWeight = sp.vstack((geneWeight, each.layers['geneWeight']), format='csr')
            if 'X_pca' in each.obsm.keys():
                pca = np.concatenate((pca,each.obsm['X_pca']),axis=0)
        each.obsp['gcn_connectivities'] = each.obsp['gcn_connectivities'].tocsr()
    
    obs = pd.concat(obs)
    variable = adatas[0].var 
    adata = anndata.AnnData(X=X,obs=obs,var=variable)
    gcn = csr_matrix((gcn_data, (row,col)), shape=(adata.X.shape[0],adata.X.shape[0]))

    if 'geneWeight' in each.layers.keys():
        adata.layers['geneWeight'] = geneWeight
    adata.obsp['gcn_connectivities'] = gcn
    if 'X_pca' in each.obsm.keys():
        adata.obsm['X_pca'] = pca
  
    return adata

    # adata0.obsp['gcn_connectivities'] = adata0.obsp['gcn_connectivities'].tocoo()
    # adata1.obsp['gcn_connectivities'] = adata1.obsp['gcn_connectivities'].tocoo()
    # adata2.obsp['gcn_connectivities'] = adata2.obsp['gcn_connectivities'].tocoo()
    # adata3.obsp['gcn_connectivities'] = adata3.obsp['gcn_connectivities'].tocoo()
    # gcn_data = adata0.obsp['gcn_connectivities'].data.tolist()
    # col = adata0.obsp['gcn_connectivities'].col.tolist()
    # row = adata0.obsp['gcn_connectivities'].row.tolist()
    # col+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].col).tolist()
    # row+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].row).tolist()
    # gcn_data+=adata1.obsp['gcn_connectivities'].data.tolist()
    # col+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].col).tolist()
    # row+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].row).tolist()
    # gcn_data+=adata2.obsp['gcn_connectivities'].data.tolist()
    # col+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].shape[0]+ adata3.obsp['gcn_connectivities'].col).tolist()
    # row+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].shape[0]+ adata3.obsp['gcn_connectivities'].row).tolist()
    # print(max(row))
    # gcn_data+=adata3.obsp['gcn_connectivities'].data.tolist()
def mergeAdata(path,total_stages):
    '''
    merge adata file from different stage to a whole one
    '''
    #read stage datasets
    adatas = []
    for i in range(total_stages):
        adata = sc.read_h5ad(os.path.join(path, 'stagedata/%d.h5ad'%i))
        adata.obs['stage'] = i
        adatas.append(adata)
    # adata0 = sc.read_h5ad(os.path.join(path, 'stagedata/0.h5ad'))
    # adata1 = sc.read_h5ad(os.path.join(path, 'stagedata/1.h5ad'))
    # adata2 = sc.read_h5ad(os.path.join(path,'stagedata/2.h5ad'))
    # adata3 = sc.read_h5ad(os.path.join(path,'stagedata/3.h5ad'))
    for i, each in enumerate(adatas):
        adatas[i].obsp['gcn_connectivities'] = adatas[i].obsp['gcn_connectivities'].tocoo()
    # adata0.obsp['gcn_connectivities'] = adata0.obsp['gcn_connectivities'].tocoo()
    # adata1.obsp['gcn_connectivities'] = adata1.obsp['gcn_connectivities'].tocoo()
    # adata2.obsp['gcn_connectivities'] = adata2.obsp['gcn_connectivities'].tocoo()
    # adata3.obsp['gcn_connectivities'] = adata3.obsp['gcn_connectivities'].tocoo()
    
    gcn_data = adatas[0].obsp['gcn_connectivities'].data.tolist()
    col = adatas[0].obsp['gcn_connectivities'].col.tolist()
    row = adatas[0].obsp['gcn_connectivities'].row.tolist()
    for i in range(1,total_stages):

        current_shape = sum([adatas[i].obsp['gcn_connectivities'].shape[0] for i in range(i)])
        col+=(current_shape+ adatas[i].obsp['gcn_connectivities'].col).tolist()
        row+=(current_shape+ adatas[i].obsp['gcn_connectivities'].row).tolist()
        gcn_data+=adatas[i].obsp['gcn_connectivities'].data.tolist()

    # gcn_data+=adata1.obsp['gcn_connectivities'].data.tolist()
    # col+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].col).tolist()
    # row+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].row).tolist()
    # gcn_data+=adata2.obsp['gcn_connectivities'].data.tolist()
    # col+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].shape[0]+ adata3.obsp['gcn_connectivities'].col).tolist()
    # row+=(adata0.obsp['gcn_connectivities'].shape[0]+ adata1.obsp['gcn_connectivities'].shape[0]+ adata2.obsp['gcn_connectivities'].shape[0]+ adata3.obsp['gcn_connectivities'].row).tolist()
    # gcn_data+=adata3.obsp['gcn_connectivities'].data.tolist()

    #get X
    if sp.isspmatrix(adatas[0].X):
        X = adatas[0].X
    else:
        X = csr_matrix(adatas[0].X)
    for i in range(1,total_stages):
        X = sp.vstack((X, adatas[i].X), format='csr')
    # X = sp.vstack((X, adata1.X), format='csr')
    # X = sp.vstack((X, adata2.X), format='csr')
    # X = sp.vstack((X, adata3.X), format='csr')
    
    #get geneWeight
    geneWeight = adatas[0].layers['geneWeight']
    for i in range(1,total_stages):
        geneWeight = sp.vstack((geneWeight, adatas[i].layers['geneWeight']), format='csr')
    # geneWeight = sp.vstack((geneWeight, adata1.layers['geneWeight']), format='csr')
    # geneWeight = sp.vstack((geneWeight, adata2.layers['geneWeight']), format='csr')
    # geneWeight = sp.vstack((geneWeight, adata3.layers['geneWeight']), format='csr')
    #get var
    if 'concat' in adatas[0].layers.keys():
        concat = adatas[0].layers['concat']
        for i in range(1,total_stages):
            concat = sp.vstack((concat, adatas[i].layers['concat']), format='csr')
        # concat = sp.vstack((concat, adata1.layers['concat']), format='csr')
        # concat = sp.vstack((concat, adata2.layers['concat']), format='csr')
        # concat = sp.vstack((concat, adata3.layers['concat']), format='csr')
    
    variable = adatas[0].var 
    
    # adata0.obs['stage'] = 0
    # adata1.obs['stage'] = 1
    # adata2.obs['stage'] = 2
    # adata3.obs['stage'] = 3
    
    #get obs
    obs = [each.obs for each in adatas]
    
    obs = pd.concat(obs)

    #get adata.ump['z']
    Z = adatas[0].obsm['z']
    for i in range(1,total_stages):
        Z = np.concatenate((Z,adatas[i].obsm['z']),axis=0)
    # Z = np.concatenate((adata0.obsm['z'],adata1.obsm['z']),axis=0)
    # Z = np.concatenate((Z,adata2.obsm['z']),axis=0)
    # Z = np.concatenate((Z,adata3.obsm['z']),axis=0)
    #get adata.ump['umap']
    umap = adatas[0].obsm['X_umap']
    for i in range(1,total_stages):
        umap = np.concatenate((umap,adatas[i].obsm['X_umap']),axis=0)
    # umap = np.concatenate((adata0.obsm['X_umap'],adata1.obsm['X_umap']),axis=0)
    # umap = np.concatenate((umap,adata2.obsm['X_umap']),axis=0)
    # umap = np.concatenate((umap,adata3.obsm['X_umap']),axis=0)
    #get top Gene
    topGene = {}
    top_gene_fold_change = {}
    top_gene_pvals_adj = {}
    clustertype = {}
    for i in range(total_stages):
        topGene[str(i)] = adatas[i].uns['topGene']
        top_gene_fold_change[str(i)]=adatas[i].uns['logfoldchanges']
        top_gene_pvals_adj[str(i)]=adatas[i].uns['top_gene_pvals_adj']
        clustertype[str(i)]=adatas[i].uns['clusterType']
    # topGene['0']=adata0.uns['topGene']
    # topGene['1']=adata1.uns['topGene']
    # topGene['2']=adata2.uns['topGene']
    # topGene['3']=adata3.uns['topGene']
    #top gene fold_change 
        adatas[i].uns['logfoldchanges'] = []
        adatas[i].uns['top_gene_pvals_adj'] = []
        for j in set(adatas[i].obs['leiden']):
            adatas[i].uns['logfoldchanges'].append(adatas[i].uns['rank_genes_groups']['logfoldchanges'][str(j)])
            adatas[i].uns['top_gene_pvals_adj'].append(adatas[i].uns['rank_genes_groups']['pvals_adj'][str(j)])

    # adata0.uns['logfoldchanges'] = []
    # adata0.uns['top_gene_pvals_adj'] = []
    # for i in set(adata0.obs['leiden']):
    #     adata0.uns['top_gene_pvals_adj'].append(adata0.uns['rank_genes_groups']['pvals_adj'][str(i)])
    #     adata0.uns['logfoldchanges'].append(adata0.uns['rank_genes_groups']['logfoldchanges'][str(i)])
    # adata1.uns['logfoldchanges'] = []
    # adata1.uns['top_gene_pvals_adj'] = []
    # for i in set(adata1.obs['leiden']):
    #     adata1.uns['top_gene_pvals_adj'].append(adata1.uns['rank_genes_groups']['pvals_adj'][str(i)])
    #     adata1.uns['logfoldchanges'].append(adata1.uns['rank_genes_groups']['logfoldchanges'][str(i)])
    # adata2.uns['logfoldchanges'] = []
    # adata2.uns['top_gene_pvals_adj'] = []
    # for i in set(adata2.obs['leiden']):
    #     adata2.uns['top_gene_pvals_adj'].append(adata2.uns['rank_genes_groups']['pvals_adj'][str(i)])
    #     adata2.uns['logfoldchanges'].append(adata2.uns['rank_genes_groups']['logfoldchanges'][str(i)])
    # adata3.uns['logfoldchanges'] = []
    # adata3.uns['top_gene_pvals_adj'] = []
    # for i in set(adata3.obs['leiden']):
    #     adata3.uns['top_gene_pvals_adj'].append(adata3.uns['rank_genes_groups']['pvals_adj'][str(i)])
    #     adata3.uns['logfoldchanges'].append(adata3.uns['rank_genes_groups']['logfoldchanges'][str(i)])
    # top_gene_fold_change = {}
    # top_gene_fold_change['0']=adata0.uns['logfoldchanges']
    # top_gene_fold_change['1']=adata1.uns['logfoldchanges']
    # top_gene_fold_change['2']=adata2.uns['logfoldchanges']
    # top_gene_fold_change['3']=adata3.uns['logfoldchanges']
    #top gene pvals_adj
    # top_gene_pvals_adj = {}
    # top_gene_pvals_adj['0']=adata0.uns['top_gene_pvals_adj']
    # top_gene_pvals_adj['1']=adata1.uns['top_gene_pvals_adj']
    # top_gene_pvals_adj['2']=adata2.uns['top_gene_pvals_adj']
    # top_gene_pvals_adj['3']=adata3.uns['top_gene_pvals_adj']
    #get uns.edges
    edges = eval(open(os.path.join(path,'edges.txt')).read())
    #get clusterType

    # clustertype = {}
    
    # clustertype['0']=adata0.uns['clusterType']
    # clustertype['1']=adata1.uns['clusterType']
    # clustertype['2']=adata2.uns['clusterType']
    # clustertype['3']=adata3.uns['clusterType']
    clusterType = clustertype
    #build new anndata and assign attribtues and write dataset
    adata = anndata.AnnData(X=X,obs=obs,var=variable)
    adata.layers['geneWeight'] = csr_matrix(geneWeight)
    # if 'concat' in adata0.layers.keys():
        # adata.layers['concat'] = csr_matrix(concat)
    adata.uns['clusterType']=clusterType
    adata.uns['edges']=edges
    adata.uns['topGene']=topGene
    adata.uns['top_gene_fold_change']=top_gene_fold_change
    adata.uns['top_gene_pvals_adj']=top_gene_pvals_adj
    adata.obsm['z']=Z
    adata.obsm['umap']=umap
    adata.obsm['X_umap'] = umap
    adata.obs['leiden'] = adata.obs['leiden'].astype(str)
    adata.obsp = {}

    gcn = csr_matrix((gcn_data, (row,col)), shape=(adata.X.shape[0],adata.X.shape[0]))
    adata.obsp['gcn_connectivities'] = gcn #np.zeros(shape=(adata.X.shape[0],adata.X.shape[0]),dtype=np.float32)
    attribute = adata.uns
    with open(os.path.join(path,'stagedata/org_attribute.pkl'),'wb') as f:
        pickle.dump(attribute, f)
    del adata.uns
    adata.write_h5ad(os.path.join(path,'stagedata/org_dataset.h5ad'),compression='gzip',compression_opts=9)



def getUnsDict(adatas,key):
    dic = {}
    for i in range(len(adatas)):
        dic[str(i)] = adatas[i].uns[key]
    # dic['0'] = adata0.uns[key]
    # dic['1'] = adata1.uns[key]
    # dic['2'] = adata2.uns[key]
    # dic['3'] = adata3.uns[key]
    return dic


# get top markers for each track from IDREM results, the requirements for top markers are:
# 1. the tendency of the marker is monotonic
# 2. the log2FC of the marker is larger than a cutoff
import json
import os
import numpy as np
import pandas as pd
from scipy.stats import norm
def scoreAgainstBackground(background,input,all=False,mean=None,std=None):
    #calculate pval for each gene
    #input is a list chages for all genes
    if all:
        background = background.reshape(-1,1)

    # print(background.mean(axis=0), background.std(axis=0))
    if mean is None:
        cdf = norm.cdf(input, loc=background.mean(axis=0), scale=background.std(axis=0))
    else:
        cdf = norm.cdf(input, loc=mean, scale=std)
    
    return cdf

def readIdremJson(path, filename):
    print('getting Target genes from ', filename)

    path = os.path.join(path,filename,'DREM.json')
    f=open(path,"r")
    lf=f.readlines()
    f.close()
    lf="".join(lf)
    lf=lf[5:-2]+']'
    tt=json.loads(lf,strict=False)
    return tt
def getTopMarkers(idrem, background,filename, cutoff=None,topN=None,one_dist=False):
    topMarkers = {}
    tt = readIdremJson(idrem,filename)

    background = np.array(background)
    
    
    temp = np.array(tt[8])
    idrem_genes = np.array(temp[1:,0].tolist())
    tendency = temp[1:,4].astype(float)* temp[1:,3].astype(float) * temp[1:,2].astype(float) * temp[1:,1].astype(float)
    tendency[tendency <0] = 0
    index = [i for i, x in enumerate(tendency) if x <= 0]
    change = temp[1:,4].astype(float) - temp[1:,1].astype(float)
    stage0 = temp[1:,1].astype(float)
    stage1 = temp[1:,2].astype(float)
    stage2 = temp[1:,3].astype(float)
    stage3 = temp[1:,4].astype(float)
    
    change[index] = 0
    if cutoff is not None:
        topMarkers['increasing'] = {}
        topMarkers['decreasing'] = {}
        increasing_stop = np.where(change > 0)[0]
        decreasing_stop = np.where(change < 0)[0]
        temp_change = change[increasing_stop]
        temp_names = idrem_genes[increasing_stop]
        temp_stage0 = stage0[increasing_stop]
        temp_stage1 = stage1[increasing_stop]
        temp_stage2 = stage2[increasing_stop]
        temp_stage3 = stage3[increasing_stop]
        temp_background = background[:,increasing_stop]

        
        increasing_stop = temp_change.argsort()[::-1]
        # increasing_pval = scoreAgainstBackground(background[:,increasing_stop],temp_change)
        
        topMarkers['increasing']['gene'] = {}
        topMarkers['increasing']['log2fc'] = {}
        topMarkers['increasing']['rank'] = {}
        topMarkers['increasing']['stage0'] = {}
        topMarkers['increasing']['stage1'] = {}
        topMarkers['increasing']['stage2'] = {}
        topMarkers['increasing']['stage3'] = {}
        topMarkers['increasing']['qval'] = {}
        count = 0
        pvals = []
        in_pvals = []
        if one_dist:
            temp_background = temp_background.reshape(-1,1)
            mean = temp_background.mean()
            std = temp_background.std()
        for i, each in enumerate(increasing_stop):
            if one_dist:
                increasing_pval = scoreAgainstBackground(temp_background,temp_change[each],mean = mean, std = std)
            else:
                increasing_pval = scoreAgainstBackground(temp_background[:, each],temp_change[each])
                if increasing_pval < (1-cutoff):
                    continue
                increasing_pval = scoreAgainstBackground(temp_background,temp_change[each],all=True)
            if increasing_pval < (1-cutoff):
                continue
            in_pvals.append(each)
            pvals.append(1-increasing_pval)
            # topMarkers['increasing']['gene'][str(count)] = temp_names[each]
            # topMarkers['increasing']['stage0'][str(count)] = temp_stage0[each]
            # topMarkers['increasing']['stage1'][str(count)] = temp_stage1[each]
            # topMarkers['increasing']['stage2'][str(count)] = temp_stage2[each]
            # topMarkers['increasing']['stage3'][str(count)] = temp_stage3[each]
            # topMarkers['increasing']['log2fc'][str(count)] = temp_change[each]
            # topMarkers['increasing']['rank'][str(count)] = count+1
            # topMarkers['increasing']['pval'][str(count)] = 1-increasing_pval#[each]
            count+=1
        count_1 = 0
        #sort pvals by sorted
        pvals = np.array(pvals).reshape(-1)

        in_pvals = np.array(in_pvals).reshape(-1)
        pvals = pvals[pvals.argsort()]

        in_pvals = in_pvals[pvals.argsort()]
        for i in range(count):
            
            q_val = pvals[i] * count / (i+1)
            if q_val > cutoff:
                continue
            each = in_pvals[i]
            topMarkers['increasing']['gene'][str(count_1)] = temp_names[each]
            topMarkers['increasing']['stage0'][str(count_1)] = temp_stage0[each]
            topMarkers['increasing']['stage1'][str(count_1)] = temp_stage1[each]
            topMarkers['increasing']['stage2'][str(count_1)] = temp_stage2[each]
            topMarkers['increasing']['stage3'][str(count_1)] = temp_stage3[each]
            topMarkers['increasing']['log2fc'][str(count_1)] = temp_change[each]
            topMarkers['increasing']['rank'][str(count_1)] = count_1+1
            topMarkers['increasing']['qval'][str(count_1)] = q_val
            count_1+=1
        temp_change = change[decreasing_stop]
        temp_names = idrem_genes[decreasing_stop]
        temp_stage0 = stage0[decreasing_stop]
        temp_stage1 = stage1[decreasing_stop]
        temp_stage2 = stage2[decreasing_stop]
        temp_stage3 = stage3[decreasing_stop]
        temp_background = background[:,decreasing_stop]
        decreasing_stop = temp_change.argsort()
        # decreasing_pval = scoreAgainstBackground(background[:,decreasing_stop],temp_change)
        
        topMarkers['decreasing']['gene'] = {}
        topMarkers['decreasing']['log2fc'] = {}
        topMarkers['decreasing']['rank'] = {}
        topMarkers['decreasing']['stage0'] = {}
        topMarkers['decreasing']['stage1'] = {}
        topMarkers['decreasing']['stage2'] = {}
        topMarkers['decreasing']['stage3'] = {}
        topMarkers['decreasing']['qval'] = {}
        count = 0
        pvals = []
        in_pvals = []
        if one_dist:
            temp_background = temp_background.reshape(-1,1)
            mean = temp_background.mean()
            std = temp_background.std()
        for i, each in enumerate(decreasing_stop):
            if one_dist:
                decreasing_pval = scoreAgainstBackground(temp_background,temp_change[each],mean = mean, std = std)
            else:
                decreasing_pval = scoreAgainstBackground(temp_background[:, each],temp_change[each])
                if decreasing_pval > cutoff:
                    continue
                decreasing_pval = scoreAgainstBackground(temp_background,temp_change[each],all=True)
            if decreasing_pval > cutoff:
                continue
            in_pvals.append(each)
            pvals.append(decreasing_pval)
            # topMarkers['decreasing']['gene'][str(count)] = temp_names[each]
            # topMarkers['decreasing']['log2fc'][str(count)] = temp_change[each]
            # topMarkers['decreasing']['rank'][str(count)] = count+1 
            # topMarkers['decreasing']['stage0'][str(count)] = temp_stage0[each]
            # topMarkers['decreasing']['stage1'][str(count)] = temp_stage1[each]
            # topMarkers['decreasing']['stage2'][str(count)] = temp_stage2[each]
            # topMarkers['decreasing']['stage3'][str(count)] = temp_stage3[each]
            # topMarkers['decreasing']['pval'][str(count)] = decreasing_pval#[each]
            count+=1
        count_1 = 0
        
        pvals = np.array(pvals).reshape(-1)
        in_pvals = np.array(in_pvals).reshape(-1)
        pvals = pvals[pvals.argsort()]
        in_pvals = in_pvals[pvals.argsort()]

        for i in range(count):
            q_val = pvals[i] * count/(i+1)  #1 - (1 - topMarkers['decreasing']['pval'][str(i)]) * (count/(i+1))
            if q_val > cutoff:
                continue
            each = in_pvals[i]
            topMarkers['decreasing']['gene'][str(count_1)] = temp_names[each]
            topMarkers['decreasing']['log2fc'][str(count_1)] = temp_change[each]
            topMarkers['decreasing']['rank'][str(count_1)] = count_1+1 
            topMarkers['decreasing']['stage0'][str(count_1)] = temp_stage0[each]
            topMarkers['decreasing']['stage1'][str(count_1)] = temp_stage1[each]
            topMarkers['decreasing']['stage2'][str(count_1)] = temp_stage2[each]
            topMarkers['decreasing']['stage3'][str(count_1)] = temp_stage3[each]
            topMarkers['decreasing']['qval'][str(count_1)] = q_val
            count_1+=1
    else: # if cutoff is not given, return ranked list
        topMarkers['increasing'] = {}
        topMarkers['decreasing'] = {}
        increasing_stop = np.where(change > 0)[0]
        decreasing_stop = np.where(change < 0)[0]
        temp_change = change[increasing_stop]
        temp_names = idrem_genes[increasing_stop]
        temp_stage0 = stage0[increasing_stop]
        temp_stage1 = stage1[increasing_stop]
        temp_stage2 = stage2[increasing_stop]
        temp_stage3 = stage3[increasing_stop]
        increasing_stop = temp_change.argsort()[::-1]
        topMarkers['increasing']['gene'] = {}
        topMarkers['increasing']['log2fc'] = {}
        topMarkers['increasing']['rank'] = {}
        topMarkers['increasing']['stage0'] = {}
        topMarkers['increasing']['stage1'] = {}
        topMarkers['increasing']['stage2'] = {}
        topMarkers['increasing']['stage3'] = {}
        if topN is not None:
            for i, each in enumerate(increasing_stop[:topN]):
                topMarkers['increasing']['gene'][str(i)] = temp_names[each]
                topMarkers['increasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['increasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['increasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['increasing']['stage3'][str(i)] = temp_stage3[each]
                topMarkers['increasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['increasing']['rank'][str(i)] = i+1
        else:
            for i, each in enumerate(increasing_stop):
                topMarkers['increasing']['gene'][str(i)] = temp_names[each]
                topMarkers['increasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['increasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['increasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['increasing']['stage3'][str(i)] = temp_stage3[each]
                topMarkers['increasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['increasing']['rank'][str(i)] = i+1
        temp_change = change[decreasing_stop]
        temp_names = idrem_genes[decreasing_stop]
        temp_stage0 = stage0[decreasing_stop]
        temp_stage1 = stage1[decreasing_stop]
        temp_stage2 = stage2[decreasing_stop]
        temp_stage3 = stage3[decreasing_stop]
        decreasing_stop = temp_change.argsort()
        topMarkers['decreasing']['gene'] = {}
        topMarkers['decreasing']['log2fc'] = {}
        topMarkers['decreasing']['rank'] = {}
        topMarkers['decreasing']['stage0'] = {}
        topMarkers['decreasing']['stage1'] = {}
        topMarkers['decreasing']['stage2'] = {}
        topMarkers['decreasing']['stage3'] = {}
        if topN is not None:
            for i, each in enumerate(decreasing_stop[:topN]):
                topMarkers['decreasing']['gene'][str(i)] = temp_names[each]
                topMarkers['decreasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['decreasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['decreasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['decreasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['decreasing']['stage3'][str(i)] = temp_stage3[each]
                topMarkers['decreasing']['rank'][str(i)] = i+1

        else:
            for i, each in enumerate(decreasing_stop):
                topMarkers['decreasing']['gene'][str(i)] = temp_names[each]
                topMarkers['decreasing']['log2fc'][str(i)] = temp_change[each]
                topMarkers['decreasing']['rank'][str(i)] = i+1 
                topMarkers['decreasing']['stage0'][str(i)] = temp_stage0[each]
                topMarkers['decreasing']['stage1'][str(i)] = temp_stage1[each]
                topMarkers['decreasing']['stage2'][str(i)] = temp_stage2[each]
                topMarkers['decreasing']['stage3'][str(i)] = temp_stage3[each]


    return topMarkers
def getTopMarkersFromIDREM(path, background,cutoff=None,topN=None,one_dist=False):
    out = {}
    filenames = os.listdir(path)
    for each in filenames:
        name = each
        if each[0] != '.':
            each = each.split('.')[0]#.split('-')[-1].split('n')
            if one_dist:
                out[each] = getTopMarkers(path,background,name,cutoff,topN,one_dist=one_dist)
            else:
                out[each] = getTopMarkers(path,background[each],name,cutoff,topN)
            # print(out[each]['decreasing'])
            # df = pd.DataFrame.from_dict(out[each], orient='index')
    return out

def runGetProgressionMarkercsv(directory,background, topN=None,cutoff=None):
    background = dict(np.load(background,allow_pickle=True).tolist())
    out = getTopMarkersFromIDREM(directory,background,topN=topN,cutoff=cutoff)
    results = pd.json_normalize(out)
    results.columns = results.columns.str.split(".").map(tuple)
    results = results.stack([0,1,2]).reset_index(0, drop=True)
    results = results.transpose()
    results = results.reset_index()
    results['index'] = results['index'].astype('int32')
    results = results.set_index(['index']).sort_index()
    # results = results.transpose()

    results = results.sort_index()
    # print(results)
    return results
    # results.to_csv('mesProgressionMarker_pval_twofilters.csv') 


def runGetProgressionMarker(directory,background, cutoff=0.05, topN=None):
    out = getTopMarkersFromIDREM(directory,background,cutoff = cutoff)
    return out
# df = pd.DataFrame.from_dict({(i,j,k): out[i][j][k]

                    #        for i in out.keys()  
                    #        for j in out[i].keys()
                    #        for k in out[i][j].keys()},
                    #    orient='index')
    # results = pd.json_normalize(out)
    # results.columns = results.columns.str.split(".").map(tuple)
    # results = results.stack([0,1,2]).reset_index(0, drop=True)
    # results = results.transpose()
    # results = results.reset_index()
    # results['index'] = results['index'].astype('int32')
    # results = results.set_index(['index']).sort_index()
    # results = results.transpose()
    
    # results = results.sort_index()
    # results.to_csv('kankan.csv') 
    # print(results.index.tolist())

def runGetProgressionMarker_one_dist(directory,background,size, cutoff=0.05, topN=None):
    one_dist  = []
    for each in background.keys(): 
        one_dist.append(np.array(background[each]))
    background = np.array(one_dist).reshape(-1, size)
    out = getTopMarkersFromIDREM(directory,background,cutoff = cutoff,one_dist=True)
    return out


#this script is used to shuffle the gene expression for each stage data 
#and and use the shuffled data to build the random background to calculate the p-val for progression markers
import scanpy as sc
import numpy as np
def averageNode(nodes,state):
    '''
    calculate the average gene expression of sibling nodes
    args: 
    nodes: number of sibling nodes
    state: the gene expression of each cluster in a certain stage
    
    return: 
    out: the average gene expression of sibling nodes
    '''
    out = 0
    for each in nodes:
        out+=state[each]
    return out/len(nodes)
def getClusterPaths(edges):
    '''
    obtain the paths of each clusters
    args:
    edeges: contains three lists of edges between control group and IPF 1 stage, IPF 1 stage and IPF 2 stage, IPF 2 stage and IPF 3 stage
    
    return:
    paths: a collection of paths of clusters
    '''
    paths = {}
    C2oneEdge = edges['0']
    one2twoEdge = edges['1']
    two2threeEdge = edges['2']
    for each in C2oneEdge:
        if str(each[0]) not in paths.keys():
            paths[str(each[0])]=[[each[0]],[each[1]]]
        else:
            paths[str(each[0])][1].append(each[1])

    #connect2 = {}
    for each in one2twoEdge:
        for item in paths.keys():
            if each[0] in paths[item][1]:
                if len(paths[item]) == 2:
                    paths[item].append([each[1]])
                else:
                    paths[item][2].append(each[1])
                

    for each in two2threeEdge:
        for item in paths.keys():
            if len(paths[item]) == 2:
                continue
            if each[0] in paths[item][2]:
                if len(paths[item]) == 3:
                    paths[item].append([each[1]])
                else:
                    paths[item][3].append(each[1])  
    return paths
def getClusterIdrem(paths,state):
    '''
    concatenate the average gene expression in a cluster tree. shape: [number of stages, number of gene]
    args: 
    paths: the collection of paths
    state: a list of average gene expression of each state
    
    return: 
    out: a list of gene expression of each cluster tree
    '''
    out = []
    for i,each in enumerate(paths.keys()):
        if len(paths[each]) == 4:
            stages = []
            for j, item in enumerate(paths[each]):
                
                stages.append(averageNode(item,state[j]))
            joint_matrix = np.concatenate((stages[0].reshape(-1,1), stages[1].reshape(-1,1), stages[2].reshape(-1,1), stages[3].reshape(-1,1)),axis =1)
            out.append(joint_matrix)
    return out
# data_dir = 'iterativeTrainingNOV26/4/stagedata/kkdataset.h5ad'
# permutation_time = 1000
# adata = sc.read(data_dir)
def get_progressionmarker_background(times, adata):
    results = {}
    edges = adata.uns['edges']
    paths = getClusterPaths(edges)
    
    stage_data =[]
    for i in range(len(adata.obs['stage'].unique())):
        adata.obs['stage'] = adata.obs['stage'].astype(str)
        temp = adata[adata.obs[adata.obs['stage']==str(i)].index.tolist()]
        stage_data.append(temp)
    for time in range(times):
        print('shuffled times:',time)
        # results.append([])
        #shuffle the gene expression for each stage data and each gene
        avg_expression_val = []
        for i in range(len(stage_data)):
            avg_expression_val.append([])
            tempX = stage_data[i].X.toarray()
          
            # for j in range(stage_data[i].shape[1]):
            #     np.random.shuffle(tempX[:,j])
            np.random.shuffle(tempX)
            stage_data[i].layers['shuffle_X'] = tempX
            for cluster in range(len(stage_data[i].obs['leiden'].unique())):
                avg_expression_val[i].append(np.mean(stage_data[i][stage_data[i].obs[stage_data[i].obs['leiden']==str(cluster)].index.tolist()].layers['shuffle_X'],axis=0))
            del stage_data[i].layers['shuffle_X']

        idrem = np.array(getClusterIdrem(paths,avg_expression_val))
        path = [each for each in paths.values() if len(each) == 4]
        for i, each in enumerate(path):
            track_name = ''
            for j, stage in enumerate(each):
                if j > 0:
                    track_name += '-'
                for k, cluster in enumerate(stage):
                    if k > 0:
                        track_name += 'n'
                    track_name += str(cluster)
            if track_name not in results.keys():
                results[track_name] = []
            results[track_name].append(idrem[i][:,-1] - idrem[i][:,0])
    return results
# out = permutation(permutation_time,adata)
# out = np.array(out).reshape(-1,adata.X.shape[1])
# print(out.shape)
# np.save('progressionMarker_background.npy',out)


import subprocess
from .gcn_utilis import  setup_graph
from .CPO_utils import get_neighbors, auto_resolution
from .utils import saveRep,get_all_adj_adata,mergeAdata,updateAttributes,get_data_file_path
from .buildGraph import getandUpadateEdges
import gc
import scanpy as sc
import numpy as np
import os
from .processIDREM import getClusterPaths, getClusterIdrem, runIdrem
from .processTFs import getTFs, getTargetGenes, matchTFandTGWithFoldChange, updataGeneTablesWithDecay
from sklearn.neighbors import kneighbors_graph
from scipy.sparse import csr_matrix
class UNAGI_runner():
    def __init__(self,data_path,total_stage,iteration,trainer,idrem_dir):
        super(UNAGI_runner, self).__init__()
        self.data_path = data_path
        self.total_stage = total_stage
        self.iteration = iteration
        self.trainer = trainer
        self.resolutions = None
        self.idrem_dir = idrem_dir
        self.neighbor_parameters = None
        self.setup_CPO = False
        self.species = None
        self.setup_IDREM = False
    def load_stage_data(self):
        stageadata = []
        for i in range(self.total_stage):
            if self.iteration != 0:
                adata = sc.read(os.path.join(self.data_path, '%d/stagedata/%d.h5ad'%(self.iteration-1,i)))
            elif self.iteration == 0:
                adata = sc.read(os.path.join(self.data_path, '%d.h5ad'%(i)))
            if 'leiden_colors' in adata.uns:
                del adata.uns['leiden_colors']
            stageadata.append(adata)
        self.all_in_one = get_all_adj_adata(stageadata)
        self.adata_stages = stageadata
        
    def annotate_stage_data(self, adata,stage):
        z_locs, z_scales, cell_embeddings = self.trainer.get_latent_representation(adata,self.iteration,self.data_path)
        adata.obsm['z'] = cell_embeddings
        if self.neighbor_parameters is None:
            sc.pp.neighbors(adata, use_rep="z",n_neighbors=50,method='umap')
            return adata
        if 'connectivities' in adata.obsp.keys():
            del adata.obsp['connectivities']
            del adata.obsp['distances']
        z_adj = kneighbors_graph(adata.obsm['z'], self.neighbor_parameters[stage], mode='connectivity', include_self=True,n_jobs=20)
        adata.obsp['connectivities'] = z_adj
        adata.obsp['distances'] = kneighbors_graph(adata.obsm['z'], self.neighbor_parameters[stage], mode='distance', include_self=True,n_jobs=20)
        sc.tl.leiden(adata,resolution = self.resolutions[stage])
        sc.tl.paga(adata)
        sc.pl.paga(adata,show=False)
        # sc.tl.umap(adata, min_dist=0.25/self.resolution_coefficient*len(adata),init_pos='paga')
        sc.tl.umap(adata, min_dist=0.05,init_pos='paga')
        rep=[z_locs, z_scales]
        adata.obs['ident'] = 'None'
        adata.obs['leiden'] = adata.obs['leiden'].astype(str)
        sc.pl.umap(adata,color='leiden',show=False)
        adata,averageValue,reps = updateAttributes(adata,rep)#add top genes, cell types attributes and get average cluster value
  
        adata.obs['leiden']=adata.obs['leiden'].astype(str)
        if self.iteration == 0:
            allzeros = np.zeros_like(adata.X)
            allzeros = csr_matrix(allzeros)
            adata.layers['geneWeight'] = allzeros
        sc.pl.umap(adata,color='ident',show=False)
        adata.write(os.path.join(self.data_path,str(self.iteration)+'/stagedata/%d.h5ad'%stage),compression='gzip',compression_opts=9)
        return adata,averageValue,reps
    
    def set_up_CPO(self, anchor_neighbors, max_neighbors, min_neighbors, resolution_min, resolution_max):
        self.setup_CPO = True
        self.anchor_neighbors = anchor_neighbors
        self.max_neighbors = max_neighbors
        self.min_neighbors = min_neighbors
        self.resolution_min = resolution_min
        self.resolution_max = resolution_max

    def run_CPO(self):
        max_adata_cells = 0
        num_cells = []
        for each in self.adata_stages:
            num_cells.append(each.shape[0])
            if len(each) > max_adata_cells:
                max_adata_cells = len(each)
        self.resolution_coefficient = max_adata_cells
        for i in range(0,len(self.adata_stages)):
            self.adata_stages[i] = self.annotate_stage_data(self.adata_stages[i], i)
        if not self.setup_CPO:
            print('CPO parameters are not set up, using default parameters')
            print('anchor_neighbors: 15, max_neighbors: 35, min_neighbors: 10, resolution_min: 0.8, resolution_max: 1.5')
            self.neighbor_parameters, anchor_index = get_neighbors(self.adata_stages, num_cells,anchor_neighbors=15,max_neighbors=35,min_neighbors=10)
            self.resolutions,_ = auto_resolution(self.adata_stages, anchor_index, self.neighbor_parameters, 0.8, 1.5)
        else:
            self.neighbor_parameters, anchor_index = get_neighbors(self.adata_stages, num_cells,anchor_neighbors=self.anchor_neighbors,max_neighbors=self.max_neighbors,min_neighbors=self.min_neighbors)
            self.resolutions,_ = auto_resolution(self.adata_stages, anchor_index, self.neighbor_parameters, self.resolution_min, self.resolution_max)
    
    def update_cell_attributes(self):
        self.averageValues = []
        reps = []
        for i in range(0,len(self.adata_stages)):
            adata = self.adata_stages[i]
            # print(adata.X.shape)
            adata.uns['topGene']={}
            adata.uns['clusterType']={}
            adata.uns['rep']={}
            adata,averageValue,rep = self.annotate_stage_data(adata,i)
            gc.collect()
            reps.append(rep)
            averageValue = np.array(averageValue)
            self.averageValues.append(averageValue)
        self.averageValues = np.array(self.averageValues,dtype=object)
        np.save(os.path.join(self.data_path, '%d/averageValues.npy'%self.iteration),self.averageValues)
        saveRep(reps,self.data_path,self.iteration)
    def build_temporal_dynamics_graph(self):
        self.edges = getandUpadateEdges(self.total_stage,self.data_path,self.iteration)
    def set_up_IDREM(self,Minimum_Absolute_Log_Ratio_Expression, Convergence_Likelihood, Minimum_Standard_Deviation):
        self.setup_IDREM = True
        self.Minimum_Absolute_Log_Ratio_Expression = Minimum_Absolute_Log_Ratio_Expression
        self.Convergence_Likelihood = Convergence_Likelihood
        self.Minimum_Standard_Deviation = Minimum_Standard_Deviation
    def register_species(self, species):
        print('Species is registered as %s'%species)
        self.species = species
    def run_IDREM(self):
        averageValues = np.load(os.path.join(self.data_path, '%d/averageValues.npy'%self.iteration),allow_pickle=True)
        paths = getClusterPaths(self.edges,self.total_stage)
        idrem= getClusterIdrem(paths,averageValues,self.total_stage)
        paths = [each for each in paths.values() if len(each) == self.total_stage]
        # print(paths) 
        idrem = np.array(idrem)
        self.genenames =np.array(list(self.adata_stages[0].var.index.values))
        if not self.setup_IDREM:
            print('IDREM parameters are not set up, using default parameters')
            print('Minimum_Absolute_Log_Ratio_Expression: 0.5, Convergence_Likelihood: 0.001, Minimum_Standard_Deviation: 0.5')
            if self.species is None:
                print('Human species is used as default')
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir)
            else:
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir,species=self.species)
        else:
            if self.species is None:
                print('Human species is used as default')
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir,Minimum_Absolute_Log_Ratio_Expression=self.Minimum_Absolute_Log_Ratio_Expression, Convergence_Likelihood=self.Convergence_Likelihood, Minimum_Standard_Deviation=self.Minimum_Standard_Deviation)
            else:
                runIdrem(paths,self.data_path,idrem,self.genenames,self.iteration,self.idrem_dir,species=self.species,Minimum_Absolute_Log_Ratio_Expression=self.Minimum_Absolute_Log_Ratio_Expression, Convergence_Likelihood=self.Convergence_Likelihood, Minimum_Standard_Deviation=self.Minimum_Standard_Deviation)
            
    def update_gene_weights_table(self,topN=100):
        TFs = getTFs(os.path.join(self.data_path,str(self.iteration)+'/'+'idremResults'+'/'))
        scope = getTargetGenes(os.path.join(self.data_path,str(self.iteration)+'/'+'idremResults'+'/'),topN)
        p = matchTFandTGWithFoldChange(TFs,scope,self.averageValues,get_data_file_path('human_encode.txt'),self.genenames)
        #np.save('../data/mes/'+str(iteration)+'/tfinfo.npy',np.array(p))
        updateLoss = updataGeneTablesWithDecay(self.data_path,str(self.iteration),p,self.total_stage)
    def build_iteration_dataset(self):
        mergeAdata(os.path.join(self.data_path,str(self.iteration)))
       
    def run(self):
        self.load_stage_data()
        if self.iteration == 0:
            is_iterative = False
        else:
            is_iterative = True
        self.trainer.train(self.all_in_one,self.iteration,target_dir=self.data_path,is_iterative=is_iterative)
        self.run_CPO()
        self.update_cell_attributes()
        self.build_temporal_dynamics_graph()
        self.run_IDREM()
        self.update_gene_weights_table()
        self.build_iteration_dataset()





import scanpy as sc
import pickle
import matplotlib.pyplot as plt
import scipy.cluster.hierarchy as sch
def plot_hc_dendrogram(adata,stage_key,celltype_key,save=False,dpi=None):
    for stage in list(adata.obs[stage_key].unique()):
        sch.dendrogram(adata.uns['hcmarkers'][str(stage)]['Z'],no_plot=True)
        #replace the labels of the leaves with the labels of the original data
        leaves = sch.leaves_list(adata.uns['hcmarkers'][str(stage)]['Z'])
        adata.obs['leiden_celltype']  = adata.obs['leiden'].astype(str) +'_'+ adata.obs[celltype_key].astype(str)
        stage0 = adata[adata.obs[stage_key] == str(stage)]
        # sc.pl.umap(stage0,color=celltype_key,show=False)
        new_leaves = []
        stage0.obs['leiden'] == stage0.obs['leiden'].astype(str)
        for each in leaves:
            temp = stage0[stage0.obs['leiden'] == str(each)]
            new_leaves.append(temp.obs['leiden_celltype'].unique()[0])
        if dpi is None:
            plt.figure(figsize=(10,10),dpi=100)
        else:
            plt.figure(figsize=(10,10),dpi=dpi)
        ax = plt.gca()
        sch.dendrogram(adata.uns['hcmarkers'][str(stage)]['Z'],ax=ax)
        ax.set_xticklabels(new_leaves)
        plt.xticks(rotation=90)
        plt.tight_layout()
        plt.title(stage)
        if save:
            plt.savefig('dendrogram_stage_%d.pdf'%(stage))
        plt.show()
if __name__ == '__main__':
    adata = sc.read_h5ad('small_1/dataset.h5ad')
    adata.uns = pickle.load(open('small_1/attribute.pkl','rb'))
    plot_hc_dendrogram(adata,'stage','ident')
